"""
microsoft_account contains functions for login with a Microsoft Account. Before using this module you need to `create a Azure application <https://docs.microsoft.com/en-us/azure/active-directory/develop/quickstart-register-app>`_.
Many thanks to wiki.vg for it's `documentation of the login process <https://wiki.vg/Microsoft_Authentication_Scheme>`_.
You may want to read the :doc:`/tutorial/microsoft_login` tutorial before using this module.
For a list of all types see :doc:`microsoft_types`.
"""
from .microsoft_types import AuthorizationTokenResponse, XBLResponse, XSTSResponse, MinecraftAuthenticateResponse, MinecraftStoreResponse, MinecraftProfileResponse, CompleteLoginResponse
from .exceptions import InvalidRefreshToken, AzureAppNotPermitted, AccountNotOwnMinecraft
from ._helper import get_user_agent, assert_func
from typing import Literal, Optional, Tuple, cast
from base64 import urlsafe_b64encode
from hashlib import sha256
import urllib.parse
import requests
import secrets

__AUTH_URL__ = "https://login.microsoftonline.com/consumers/oauth2/v2.0/authorize"
__TOKEN_URL__ = "https://login.microsoftonline.com/consumers/oauth2/v2.0/token"
__SCOPE__ = "XboxLive.signin offline_access"


def get_login_url(client_id: str, redirect_uri: str) -> str:
    """
    Generate a login url.\\
    For a more secure alternative, use :func:`get_secure_login_data`

    :param client_id: The Client ID of your Azure App
    :param redirect_uri: The Redirect URI of your Azure App
    :return: The url to the website on which the user logs in
    """
    parameters = {
        "client_id": client_id,
        "response_type": "code",
        "redirect_uri": redirect_uri,
        "response_mode": "query",
        "scope": __SCOPE__,
    }

    url = urllib.parse.urlparse(__AUTH_URL__)._replace(query=urllib.parse.urlencode(parameters)).geturl()
    return url


def _generate_pkce_data() -> Tuple[str, str, Literal["plain", "S256"]]:
    """
    Generates the PKCE code challenge and code verifier

    :return: A tuple containing the code_verifier, the code_challenge and the code_challenge_method.
    """
    code_verifier = secrets.token_urlsafe(128)[:128]
    code_challenge = urlsafe_b64encode(sha256(code_verifier.encode("ascii")).digest()).decode("ascii")[:-1]
    code_challenge_method = "S256"
    return code_verifier, code_challenge, cast(Literal["plain", "S256"], code_challenge_method)


def generate_state() -> str:
    """
    Generates a random state
    """
    return secrets.token_urlsafe(16)


def get_secure_login_data(client_id: str, redirect_uri: str, state: Optional[str] = None) -> Tuple[str, str, str]:
    """
    Generates the login data for a secure login with pkce and state.\\
    Prevents Cross-Site Request Forgery attacks and authorization code injection attacks.

    :param client_id: The Client ID of your Azure App
    :param redirect_uri: The Redirect URI of your Azure App
    :param state: You can use a existing state. If not set, a state will be generated using :func:`generate_state`.
    """
    code_verifier, code_challenge, code_challenge_method = _generate_pkce_data()

    if state is None:
        state = generate_state()

    parameters = {
        "client_id": client_id,
        "response_type": "code",
        "redirect_uri": redirect_uri,
        "response_mode": "query",
        "scope": __SCOPE__,
        "state": state,
        "code_challenge": code_challenge,
        "code_challenge_method": code_challenge_method
    }

    url = urllib.parse.urlparse(__AUTH_URL__)._replace(query=urllib.parse.urlencode(parameters)).geturl()

    return url, state, code_verifier


def url_contains_auth_code(url: str) -> bool:
    """
    Checks if the given url contains a authorization code

    :param url: The URL to check
    """
    parsed = urllib.parse.urlparse(url)
    qs = urllib.parse.parse_qs(parsed.query)
    return "code" in qs


def get_auth_code_from_url(url: str) -> Optional[str]:
    """
    Get the authorization code from the url.\\
    If you want to check the state, use :func:`parse_auth_code_url`, which throws errors instead of returning an optional value.

    :param url: The URL to parse
    :return: The auth code or None if the the code is nonexistent
    """
    parsed = urllib.parse.urlparse(url)
    qs = urllib.parse.parse_qs(parsed.query)
    try:
        return qs["code"][0]
    except KeyError:
        return None


def parse_auth_code_url(url: str, state: Optional[str]) -> str:
    """
    Parse the authorization code url and checks the state.

    :param url: The URL to parse
    :param state: If set, the function raises a AssertionError, if the state do no match the state in the URL
    :return: The auth code
    """
    parsed = urllib.parse.urlparse(url)
    qs = urllib.parse.parse_qs(parsed.query)

    if state is not None:
        assert_func(state == qs["state"][0])

    return qs["code"][0]


def get_authorization_token(client_id: str, client_secret: Optional[str], redirect_uri: str, auth_code: str, code_verifier: Optional[str] = None) -> AuthorizationTokenResponse:
    """
    Get the authorization token. This function is called during :func:`complete_login`, so you need to use this function ony if :func:`complete_login` doesnt't work for you.

    :param client_id: The Client ID of your Azure App
    :param client_secret: The Client Secret of your Azure App. This is deprecated and should not been used anymore.
    :param redirect_uri: The Redirect URI of your Azure App
    :param auth_code: The Code you get from :func:`parse_auth_code_url`
    :param code_verifier: The 3rd entry in the Tuple you get from :func:`get_secure_login_data`
    """
    parameters = {
        "client_id": client_id,
        "scope": __SCOPE__,
        "code": auth_code,
        "redirect_uri": redirect_uri,
        "grant_type": "authorization_code",
    }

    if client_secret is not None:
        parameters["client_secret"] = client_secret

    if code_verifier is not None:
        parameters["code_verifier"] = code_verifier

    header = {
        "Content-Type": "application/x-www-form-urlencoded",
        "user-agent": get_user_agent()
    }
    r = requests.post(__TOKEN_URL__, data=parameters, headers=header)
    return r.json()


def refresh_authorization_token(client_id: str, client_secret: Optional[str], redirect_uri: Optional[str], refresh_token: str) -> AuthorizationTokenResponse:
    """
    Refresh the authorization token. This function is called during :func:`complete_refresh`, so you need to use this function ony if :func:`complete_refresh` doesnt't work for you.

    :param client_id: The Client ID of your Azure App
    :param client_secret: The Client Secret of your Azure App. This is deprecated and should not been used anymore.
    :param redirect_uri: The Redirect URI of Azure App. This Parameter only exists for backwards compatibility and is not used anymore.
    :param refresh_token: Your refresh token
    """
    parameters = {
        "client_id": client_id,
        "scope": __SCOPE__,
        "refresh_token": refresh_token,
        "grant_type": "refresh_token"
    }

    if client_secret is not None:
        parameters["client_secret"] = client_secret

    # redirect_uri was used in a previous version of this library
    # we keep it for backwards compatibility, but it is not required anymore
    _ = redirect_uri

    header = {
        "user-agent": get_user_agent()
    }
    r = requests.post("https://login.live.com/oauth20_token.srf", data=parameters, headers=header)
    return r.json()


def authenticate_with_xbl(access_token: str) -> XBLResponse:
    """
    Authenticate with Xbox Live. This function is called during :func:`complete_login`, so you need to use this function ony if :func:`complete_login` doesnt't work for you.

    :param access_token: The Token you get from :func:`get_authorization_token`
    """
    parameters = {
        "Properties": {
            "AuthMethod": "RPS",
            "SiteName": "user.auth.xboxlive.com",
            "RpsTicket": f"d={access_token}"
        },
        "RelyingParty": "http://auth.xboxlive.com",
        "TokenType": "JWT"
    }
    header = {
        "Content-Type": "application/json",
        "user-agent": get_user_agent(),
        "Accept": "application/json"
    }
    r = requests.post("https://user.auth.xboxlive.com/user/authenticate", json=parameters, headers=header)
    return r.json()


def authenticate_with_xsts(xbl_token: str) -> XSTSResponse:
    """
    Authenticate with XSTS. This function is called during :func:`complete_login`, so you need to use this function ony if :func:`complete_login` doesnt't work for you.

    :param xbl_token: The Token you get from :func:`authenticate_with_xbl`
    """
    parameters = {
        "Properties": {
            "SandboxId": "RETAIL",
            "UserTokens": [
                xbl_token
            ]
        },
        "RelyingParty": "rp://api.minecraftservices.com/",
        "TokenType": "JWT"
    }
    header = {
        "Content-Type": "application/json",
        "user-agent": get_user_agent(),
        "Accept": "application/json"
    }
    r = requests.post("https://xsts.auth.xboxlive.com/xsts/authorize", json=parameters, headers=header)
    return r.json()


def authenticate_with_minecraft(userhash: str, xsts_token: str) -> MinecraftAuthenticateResponse:
    """
    Authenticate with Minecraft. This function is called during :func:`complete_login`, so you need to use this function ony if :func:`complete_login` doesnt't work for you.

    :param userhash: The Hash you get from :func:`authenticate_with_xbl`
    :param xsts_token: The Token you get from :func:`authenticate_with_xsts`
    """
    parameters = {
        "identityToken": f"XBL3.0 x={userhash};{xsts_token}"
    }
    header = {
        "Content-Type": "application/json",
        "user-agent": get_user_agent(),
        "Accept": "application/json"
    }
    r = requests.post("https://api.minecraftservices.com/authentication/login_with_xbox", json=parameters, headers=header)
    return r.json()


def get_store_information(access_token: str) -> MinecraftStoreResponse:
    """
    Get the store information.

    :param access_token: The Token you get from :func:`authenticate_with_minecraft`
    """
    header = {
        "Authorization": f"Bearer {access_token}",
        "user-agent": get_user_agent()
    }
    r = requests.get("https://api.minecraftservices.com/entitlements/mcstore", headers=header)
    return r.json()


def get_profile(access_token: str) -> MinecraftProfileResponse:
    """
    Get the profile. This function is called during :func:`complete_login`, so you need to use this function ony if :func:`complete_login` doesnt't work for you.

    :param access_token: The Token you get from :func:`authenticate_with_minecraft`
    """
    header = {
        "Authorization": f"Bearer {access_token}",
        "user-agent": get_user_agent()
    }
    r = requests.get("https://api.minecraftservices.com/minecraft/profile", headers=header)
    return r.json()


def complete_login(client_id: str, client_secret: Optional[str], redirect_uri: str, auth_code: str, code_verifier: Optional[str] = None) -> CompleteLoginResponse:
    """
    Do the complete login process.

    :param client_id: The Client ID of your Azure App
    :param client_secret: The Client Secret of your Azure App. This is deprecated and should not been used anymore.
    :param redirect_uri: The Redirect URI of your Azure App
    :param auth_code: The Code you get from :func:`parse_auth_code_url`
    :param code_verifier: The 3rd entry in the Tuple you get from :func:`get_secure_login_data`
    :raises AzureAppNotPermitted: Your Azure App don't have the Permission to use the Minecraft API
    :raises AccountNotOwnMinecraft: The Account does not own Minecraft

    It returns the following:

    .. code:: json

        {
            "id" : "The uuid",
            "name" : "The username",
            "access_token": "The acces token",
            "refresh_token": "The refresh token",
            "skins" : [{
                "id" : "6a6e65e5-76dd-4c3c-a625-162924514568",
                "state" : "ACTIVE",
                "url" : "http://textures.minecraft.net/texture/1a4af718455d4aab528e7a61f86fa25e6a369d1768dcb13f7df319a713eb810b",
                "variant" : "CLASSIC",
                "alias" : "STEVE"
            } ],
            "capes" : []
        }
    """
    token_request = get_authorization_token(client_id, client_secret, redirect_uri, auth_code, code_verifier)
    token = token_request["access_token"]

    xbl_request = authenticate_with_xbl(token)
    xbl_token = xbl_request["Token"]
    userhash = xbl_request["DisplayClaims"]["xui"][0]["uhs"]

    xsts_request = authenticate_with_xsts(xbl_token)
    xsts_token = xsts_request["Token"]

    account_request = authenticate_with_minecraft(userhash, xsts_token)

    if "access_token" not in account_request:
        raise AzureAppNotPermitted()

    access_token = account_request["access_token"]

    profile = get_profile(access_token)

    if "error" in profile and profile["error"] == "NOT_FOUND":
        raise AccountNotOwnMinecraft()

    profile = cast(CompleteLoginResponse, profile)

    profile["access_token"] = account_request["access_token"]
    profile["refresh_token"] = token_request["refresh_token"]

    return profile


def complete_refresh(client_id: str, client_secret: Optional[str], redirect_uri: Optional[str], refresh_token: str) -> CompleteLoginResponse:
    """
    Do the complete login process with a refresh token. It returns the same as :func:`complete_login`.

    :param client_id: The Client ID of your Azure App
    :param client_secret: The Client Secret of your Azure App. This is deprecated and should not been used anymore.
    :param redirect_uri: The Redirect URI of Azure App. This Parameter only exists for backwards compatibility and is not used anymore.
    :param refresh_token: Your refresh token
    :raises InvalidRefreshToken: Your refresh token is not valid
    :raises AccountNotOwnMinecraft: The Account does not own Minecraft
    """
    token_request = refresh_authorization_token(client_id, client_secret, redirect_uri, refresh_token)

    if "error" in token_request:
        raise InvalidRefreshToken()

    token = token_request["access_token"]

    xbl_request = authenticate_with_xbl(token)
    xbl_token = xbl_request["Token"]
    userhash = xbl_request["DisplayClaims"]["xui"][0]["uhs"]

    xsts_request = authenticate_with_xsts(xbl_token)
    xsts_token = xsts_request["Token"]

    account_request = authenticate_with_minecraft(userhash, xsts_token)
    access_token = account_request["access_token"]

    profile = get_profile(access_token)

    if "error" in profile and profile["error"] == "NOT_FOUND":
        raise AccountNotOwnMinecraft()

    profile = cast(CompleteLoginResponse, profile)

    profile["access_token"] = account_request["access_token"]
    profile["refresh_token"] = token_request["refresh_token"]

    return profile



import minecraft_launcher_lib
import pathlib
import shutil
import pytest
import json
import os


def _prepare_vanilla_launcher_test_env(tmp_path: pathlib.Path) -> None:
    shutil.copytree(os.path.join(os.path.dirname(__file__), "data", "vanilla_launcher"), tmp_path / ".minecraft")


def test_load_vanilla_launcher_profiles(tmp_path: pathlib.Path) -> None:
    _prepare_vanilla_launcher_test_env(tmp_path)

    profile_list = minecraft_launcher_lib.vanilla_launcher.load_vanilla_launcher_profiles(tmp_path / ".minecraft")

    assert len(profile_list) == 4

    assert profile_list[0]["name"] == "Test"
    assert profile_list[0]["version"] is None
    assert profile_list[0]["versionType"] == "latest-release"
    assert profile_list[0]["gameDirectory"] is None
    assert profile_list[0]["javaExecutable"] is None
    assert profile_list[0]["javaArguments"] == ["-Xmx2G", "-XX:+UnlockExperimentalVMOptions", "-XX:+UseG1GC", "-XX:G1NewSizePercent=20", "-XX:G1ReservePercent=20", "-XX:MaxGCPauseMillis=50", "-XX:G1HeapRegionSize=32M"]
    assert profile_list[0]["customResolution"]["height"] == 400
    assert profile_list[0]["customResolution"]["width"] == 450

    assert profile_list[1]["name"] == "Forge1.16.5"
    assert profile_list[1]["version"] == "1.16.5-forge-36.1.24"
    assert profile_list[1]["versionType"] == "custom"
    assert profile_list[1]["gameDirectory"] == "/gametest"
    assert profile_list[1]["javaExecutable"] == "/javatest"
    assert profile_list[1]["javaArguments"] is None
    assert profile_list[1]["customResolution"] is None

    assert profile_list[2]["name"] == "Latest release"
    assert profile_list[2]["version"] is None
    assert profile_list[2]["versionType"] == "latest-release"
    assert profile_list[2]["gameDirectory"] is None
    assert profile_list[2]["javaExecutable"] is None
    assert profile_list[2]["javaArguments"] is None
    assert profile_list[2]["customResolution"] is None

    assert profile_list[3]["name"] == "Latest snapshot"
    assert profile_list[3]["version"] is None
    assert profile_list[3]["versionType"] == "latest-snapshot"
    assert profile_list[3]["gameDirectory"] is None
    assert profile_list[3]["javaExecutable"] is None
    assert profile_list[3]["javaArguments"] is None
    assert profile_list[3]["customResolution"] is None


def test_vanilla_launcher_profile_to_minecraft_options() -> None:
    test_values = {"name": "test", "versionType": "latest-release"}
    assert len(minecraft_launcher_lib.vanilla_launcher.vanilla_launcher_profile_to_minecraft_options(test_values)) == 0
    assert minecraft_launcher_lib.vanilla_launcher.vanilla_launcher_profile_to_minecraft_options({**test_values, **{"gameDirectory": "test"}}) == {"gameDirectory": "test"}
    assert minecraft_launcher_lib.vanilla_launcher.vanilla_launcher_profile_to_minecraft_options({**test_values, **{"javaExecutable": "/testJava"}}) == {"executablePath": "/testJava"}
    assert minecraft_launcher_lib.vanilla_launcher.vanilla_launcher_profile_to_minecraft_options({**test_values, **{"javaArguments": ["a", "b"]}}) == {"jvmArguments": ["a", "b"]}
    assert minecraft_launcher_lib.vanilla_launcher.vanilla_launcher_profile_to_minecraft_options({**test_values, **{"customResolution": {"height": 100, "width": 150}}}) == {
        "customResolution": True,
        "resolutionWidth": "150",
        "resolutionHeight": "100"
    }

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.vanilla_launcher_profile_to_minecraft_options({})


def test_get_vanilla_launcher_profile_version() -> None:
    latest_version = minecraft_launcher_lib.utils.get_latest_version()

    assert minecraft_launcher_lib.vanilla_launcher.get_vanilla_launcher_profile_version({"name": "test", "version": "test", "versionType": "latest-release"}) == latest_version["release"]
    assert minecraft_launcher_lib.vanilla_launcher.get_vanilla_launcher_profile_version({"name": "test", "version": "test", "versionType": "latest-snapshot"}) == latest_version["snapshot"]
    assert minecraft_launcher_lib.vanilla_launcher.get_vanilla_launcher_profile_version({"name": "test", "version": "test", "versionType": "custom"}) == "test"

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.get_vanilla_launcher_profile_version({"name": "test", "version": "test", "versionType": "test"})


def _remove_none_values(none_dict: dict) -> dict:
    new_dict = {}
    for key, value in none_dict.items():
        if value is not None:
            new_dict[key] = value
    return new_dict


def _check_vanilla_profile_written(tmp_path: pathlib.Path, profile: minecraft_launcher_lib.types.VanillaLauncherProfile) -> None:
    temp_dir_count = 0
    while True:
        if not (tmp_path / str(temp_dir_count)).is_dir():
            break
        else:
            temp_dir_count += 1

    current_dir = (tmp_path / str(temp_dir_count))
    current_dir.mkdir()

    with open(os.path.join(os.path.dirname(__file__), "data", "vanilla_launcher", "launcher_profiles.json"), "r", encoding="utf-8") as f:
        json_data = json.load(f)

    json_data["profiles"] = {}

    with open(current_dir / "launcher_profiles.json", "w", encoding="utf-8") as f:
        json.dump(json_data, f, ensure_ascii=False)

    minecraft_launcher_lib.vanilla_launcher.add_vanilla_launcher_profile(current_dir, profile)
    assert _remove_none_values(minecraft_launcher_lib.vanilla_launcher.load_vanilla_launcher_profiles(current_dir)[0]) == _remove_none_values(profile)


def test_add_vanilla_launcher_profile(tmp_path: pathlib.Path) -> None:
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "latest-release"})
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "latest-snapshot"})
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "custom", "version": "test"})
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "latest-release", "gameDirectory": "testGame"})
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "latest-release", "javaExecutable": "testJava"})
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "latest-release", "javaArguments": ["a", "b", "c"]})
    _check_vanilla_profile_written(tmp_path, {"name": "test", "versionType": "latest-release", "customResolution": {"width": 150, "height": 100}})

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.add_vanilla_launcher_profile(tmp_path, {"name": "test", "versionType": "custom"})

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.add_vanilla_launcher_profile(tmp_path, {"name": "test", "versionType": "latest-release", "gameDirectory": 123})

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.add_vanilla_launcher_profile(tmp_path, {"name": "test", "versionType": "latest-release", "javaExecutable": 123})

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.add_vanilla_launcher_profile(tmp_path, {"name": "test", "versionType": "latest-release", "javaArguments": [123]})

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile):
        minecraft_launcher_lib.vanilla_launcher.add_vanilla_launcher_profile(tmp_path, {"name": "test", "versionType": "latest-release", "customResolution": {"width": "abc"}})


def test_do_vanilla_launcher_profiles_exists(tmp_path: pathlib.Path) -> None:
    assert minecraft_launcher_lib.vanilla_launcher.do_vanilla_launcher_profiles_exists(tmp_path / ".minecraft") is False
    _prepare_vanilla_launcher_test_env(tmp_path)
    assert minecraft_launcher_lib.vanilla_launcher.do_vanilla_launcher_profiles_exists(tmp_path / ".minecraft") is True



import minecraft_launcher_lib


def test_VersionNotFound() -> None:
    v = minecraft_launcher_lib.exceptions.VersionNotFound("test")
    assert v.version == "test"
    assert v.msg == "Version test was not found"
    assert isinstance(v, ValueError)


def test_UnsupportedVersion() -> None:
    v = minecraft_launcher_lib.exceptions.UnsupportedVersion("test")
    assert v.version == "test"
    assert v.msg == "Version test is not supported"
    assert isinstance(v, ValueError)


def test_ExternalProgramError() -> None:
    c = minecraft_launcher_lib.exceptions.ExternalProgramError(["a", "b"], "out", "err")
    assert c.command == ["a", "b"]
    assert c.stdout == "out"
    assert c.stderr == "err"
    assert isinstance(c, Exception)


def test_InvalidRefreshToken() -> None:
    i = minecraft_launcher_lib.exceptions.InvalidRefreshToken()
    assert isinstance(i, ValueError)


def test_InvalidVanillaLauncherProfile() -> None:
    ex = minecraft_launcher_lib.exceptions.InvalidVanillaLauncherProfile({"name": "test"})
    assert ex.profile == {"name": "test"}
    assert isinstance(ex, ValueError)


def test_SecurityError() -> None:
    ex = minecraft_launcher_lib.exceptions.SecurityError("code", "msg")
    assert ex.code == "code"
    assert ex.message == "msg"
    assert isinstance(ex, Exception)


def test_FileOutsideMinecraftDirectory() -> None:
    ex = minecraft_launcher_lib.exceptions.FileOutsideMinecraftDirectory("path", "dir")
    assert ex.path == "path"
    assert ex.minecraft_directory == "dir"
    assert isinstance(ex, minecraft_launcher_lib.exceptions.SecurityError)


def test_InvalidChecksum() -> None:
    ex = minecraft_launcher_lib.exceptions.InvalidChecksum("url", "path", "expected", "actual")
    assert ex.url == "url"
    assert ex.path == "path"
    assert ex.expected_checksum == "expected"
    assert ex.actual_checksum == "actual"
    assert isinstance(ex, minecraft_launcher_lib.exceptions.SecurityError)


def test_AzureAppNotPermitted() -> None:
    ex = minecraft_launcher_lib.exceptions.AzureAppNotPermitted()
    assert ex.args[0] == "It looks like your Azure App don't have the Permission to use the Minecraft API. Take a look at the Documentation for more Information."


def test_PlatformNotSupported() -> None:
    ex = minecraft_launcher_lib.exceptions.PlatformNotSupported()
    assert ex.args[0] == "Your Platform is not supported"



from typing import List, Dict, Literal, TypedDict


class _RuntimeListJsonEntryManifest(TypedDict):
    sha1: str
    size: int
    url: str


class _RuntimeListJsonEntry(TypedDict):
    availability: Dict[Literal["group", "progress"], int]
    manifest: _RuntimeListJsonEntryManifest
    version: Dict[Literal["name", "released"], str]


RuntimeListJson = Dict[str, Dict[str, List[_RuntimeListJsonEntry]]]


class _PlatformManifestJsonFileDownloads(TypedDict):
    sha1: str
    size: int
    url: str


class _PlatformManifestJsonFile(TypedDict, total=False):
    downloads: Dict[Literal["lzma", "raw"], _PlatformManifestJsonFileDownloads]
    type: Literal["file", "directory", "link"]
    executable: bool
    target: str


class PlatformManifestJson(TypedDict):
    files: Dict[str, _PlatformManifestJsonFile]



from ._test_helper import prepare_requests_mock
import minecraft_launcher_lib
import pytest_subtests
import requests_mock
import subprocess
import platform
import datetime
import pathlib
import pytest
import os


def test_get_jvm_runtimes() -> None:
    runtime_list = minecraft_launcher_lib.runtime.get_jvm_runtimes()
    for i in runtime_list:
        assert isinstance(i, str)


def test_get_installed_jvm_runtimes(tmp_path: pathlib.Path) -> None:
    assert len(minecraft_launcher_lib.runtime.get_installed_jvm_runtimes(tmp_path)) == 0

    for i in ["a", "b", "c"]:
        os.makedirs(os.path.join(tmp_path, "runtime", i))

    runtime_list = minecraft_launcher_lib.runtime.get_installed_jvm_runtimes(tmp_path)
    runtime_list.sort()
    assert runtime_list == ["a", "b", "c"]


def test_install_jvm_runtime(monkeypatch: pytest.MonkeyPatch, subtests: pytest_subtests.SubTests, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "architecture", lambda: ("test",))
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_requests_mock(requests_mock)

    with subtests.test("Empty"):
        minecraft_launcher_lib.runtime.install_jvm_runtime("java-runtime-empty", tmp_path)

        assert len(os.listdir(tmp_path)) == 0

    with subtests.test("Full"):
        # This functions needs to be patched, so the test also works on non Unix Systems
        monkeypatch.setattr(os, "symlink", lambda target, file: pathlib.Path(file).write_text(target), raising=False)
        monkeypatch.setattr(subprocess, "run", lambda cmd: pathlib.Path(cmd[2] + ".execute").touch() if cmd[0] == "chmod" else None)

        minecraft_launcher_lib.runtime.install_jvm_runtime("java-runtime-test", tmp_path)

        assert (tmp_path / "runtime" / "java-runtime-test" / "linux" / ".version").read_text().strip() == "17.0.3"
        assert os.path.isfile(tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test.sha1")
        assert os.path.isfile(tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "jre.bundle" / "Contents" / "Home" / "COPYRIGHT")
        assert os.path.isfile(tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "jre.bundle" / "Contents" / "Home" / "LICENSE")
        assert os.path.isfile(tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "jre.bundle" / "Contents" / "Home" / "LICENSE.execute")
        assert os.path.isdir(tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "lib" / "desktop" / "mime")
        assert (tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "links" / "testlink").read_text().strip() == "../target"

    with subtests.test("Invalid"):
        with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
            minecraft_launcher_lib.runtime.install_jvm_runtime("java-runtime-invalid", tmp_path)

    with subtests.test("Exceptions"):
        # Test, if the exceptions are handeld
        monkeypatch.setattr(subprocess, "run", lambda cmd: exec("raise FileNotFoundError()"))
        monkeypatch.setattr(os, "symlink", lambda target, file: exec("raise Exception()"), raising=False)
        minecraft_launcher_lib.runtime.install_jvm_runtime("java-runtime-test", tmp_path)

        with pytest.raises(minecraft_launcher_lib.exceptions.FileOutsideMinecraftDirectory):
            minecraft_launcher_lib.runtime.install_jvm_runtime("java-runtime-error", tmp_path)


def test_get_executable_path(monkeypatch: pytest.MonkeyPatch, subtests: pytest_subtests.SubTests, tmp_path: pathlib.Path) -> None:
    with subtests.test("Linux"):
        monkeypatch.setattr(platform, "architecture", lambda: ("test",))
        monkeypatch.setattr(platform, "system", lambda: "Linux")
        linux_java = (tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "bin" / "java")
        os.makedirs(linux_java.parent)
        linux_java.touch()
        assert minecraft_launcher_lib.runtime.get_executable_path("java-runtime-test", tmp_path) == str(linux_java)

    with subtests.test("Windows"):
        monkeypatch.undo()
        monkeypatch.setattr(platform, "architecture", lambda: ("test",))
        monkeypatch.setattr(platform, "system", lambda: "Windows")
        windows_java = (tmp_path / "runtime" / "java-runtime-test" / "windows-x64" / "java-runtime-test" / "bin" / "java.exe")
        os.makedirs(windows_java.parent)
        windows_java.touch()
        assert minecraft_launcher_lib.runtime.get_executable_path("java-runtime-test", tmp_path) == str(windows_java)

    with subtests.test("macOS"):
        monkeypatch.undo()
        monkeypatch.setattr(platform, "machine", lambda: "arm64")
        monkeypatch.setattr(platform, "system", lambda: "Darwin")
        mac_java = (tmp_path / "runtime" / "java-runtime-test" / "mac-os-arm64" / "java-runtime-test" / "jre.bundle" / "Contents" / "Home" / "bin" / "java")
        os.makedirs(mac_java.parent)
        mac_java.touch()
        assert minecraft_launcher_lib.runtime.get_executable_path("java-runtime-test", tmp_path) == str(mac_java)

    with subtests.test("Invalid"):
        monkeypatch.undo()
        assert minecraft_launcher_lib.runtime.get_executable_path("java-runtime-invalid", tmp_path) is None


def _test_runtime_information() -> None:
    info = minecraft_launcher_lib.runtime.get_jvm_runtime_information("java-runtime-gamma")
    assert isinstance(info["released"], datetime.datetime)
    assert isinstance(info["name"], str)
    assert len(info) == 2


def test_get_jvm_runtime_information(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(platform, "architecture", lambda: ("32bit",))
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    _test_runtime_information()

    monkeypatch.undo()
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    _test_runtime_information()

    monkeypatch.setattr(platform, "architecture", lambda: ("32bit",))
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    with pytest.raises(minecraft_launcher_lib.exceptions.PlatformNotSupported):
        minecraft_launcher_lib.runtime.get_jvm_runtime_information("java-runtime-gamma")

    monkeypatch.undo()
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    _test_runtime_information()

    monkeypatch.undo()
    monkeypatch.setattr(platform, "machine", lambda: "arm64")
    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    _test_runtime_information()

    monkeypatch.undo()
    monkeypatch.setattr(platform, "machine", lambda: "test")
    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    _test_runtime_information()

    monkeypatch.undo()
    monkeypatch.setattr(platform, "system", lambda: "Haiku")
    with pytest.raises(minecraft_launcher_lib.exceptions.PlatformNotSupported):
        minecraft_launcher_lib.runtime.get_jvm_runtime_information("java-runtime-gamma")

    monkeypatch.undo()
    with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
        minecraft_launcher_lib.runtime.get_jvm_runtime_information("InvalidVersion")



import minecraft_launcher_lib
from typing import Any
import requests
import pytest


class MicrosoftRequestsResponseMock:
    def __init__(self, code, data: Any) -> None:
        self._data = data
        self.code = code

    def json(self) -> Any:
        return self._data


class MicrosoftRequestsMock:
    def __init__(self, mode: str) -> None:
        self._mode = mode

    def get(self, url: str, **kwargs: dict) -> MicrosoftRequestsResponseMock:
        if url == "https://api.minecraftservices.com/minecraft/profile":
            if kwargs["headers"]["Authorization"] != "Bearer minecraft_access_token":
                return MicrosoftRequestsResponseMock(500, {})

            if self._mode == "not_own_minecraft":
                return MicrosoftRequestsResponseMock(400, {"error": "NOT_FOUND", "errorMessage": "Not Found"})

            return MicrosoftRequestsResponseMock(200, {"id": "minecraft_uuid", "name": "minecraft_name", "skins": [], "capes": []})

        elif url == "https://api.minecraftservices.com/entitlements/mcstore":
            if kwargs["headers"]["Authorization"] != "Bearer minecraft_access_token":
                return MicrosoftRequestsResponseMock(500, {})

            return MicrosoftRequestsResponseMock(200, {"items": [], "signature": "jwt sig", "keyId": "1"})
        else:
            raise ValueError(f"Invalid GET URL {url}")

    def post(self, url: str, **kwargs: dict) -> MicrosoftRequestsResponseMock:
        if url == "https://login.microsoftonline.com/consumers/oauth2/v2.0/token":
            return MicrosoftRequestsResponseMock(200, {"access_token": "test_access_token", "token_type": "Bearer", "expires_in": 0, "scope": "", "refresh_token": "refresh_token"})

        elif url == "https://login.live.com/oauth20_token.srf":
            if kwargs["data"]["refresh_token"] != "refresh_token":
                return MicrosoftRequestsResponseMock(403, {"error": ""})

            return MicrosoftRequestsResponseMock(200, {"access_token": "test_access_token", "token_type": "Bearer", "expires_in": 0, "scope": "", "refresh_token": "refresh_token"})

        elif url == "https://user.auth.xboxlive.com/user/authenticate":
            if kwargs["json"]["Properties"]["RpsTicket"] != "d=test_access_token":
                return MicrosoftRequestsResponseMock(500, {})

            return MicrosoftRequestsResponseMock(200, {"IssueInstant": "", "NotAfter": "", "Token": "xbl_token", "DisplayClaims": {"xui": [{"uhs": "userhash"}]}})

        elif url == "https://xsts.auth.xboxlive.com/xsts/authorize":
            if kwargs["json"]["Properties"]["UserTokens"][0] != "xbl_token":
                return MicrosoftRequestsResponseMock(500, {})

            return MicrosoftRequestsResponseMock(200, {"IssueInstant": "", "NotAfter": "", "Token": "xsts_token", "DisplayClaims": {"xui": [{"uhs": "userhash"}]}})

        elif url == "https://api.minecraftservices.com/authentication/login_with_xbox":
            if kwargs["json"]["identityToken"] != "XBL3.0 x=userhash;xsts_token":
                return MicrosoftRequestsResponseMock(500, {})

            if self._mode == "not_permitted_app":
                return MicrosoftRequestsResponseMock(403, {"error": "FORBIDDEN", "path": "/authentication/login_with_xbox"})

            return MicrosoftRequestsResponseMock(200, {"username": "", "roles": [], "access_token": "minecraft_access_token", "token_type": "Bearer", "expires_in": 0})

        else:
            raise ValueError(f"Invalid POST URL {url}")


def test_get_login_url() -> None:
    login_url = minecraft_launcher_lib.microsoft_account.get_login_url("CLIENT_ID", "REDIRECT_URL")
    assert login_url is not None


def test_generate_state() -> None:
    assert isinstance(minecraft_launcher_lib.microsoft_account.generate_state(), str)


def test_get_secure_login_data() -> None:
    login_url, state, code_verifier = minecraft_launcher_lib.microsoft_account.get_secure_login_data("CLIENT_ID", "REDIRECT_URL")

    assert login_url is not None
    assert state is not None
    assert code_verifier is not None


def test_url_contains_auth_code() -> None:
    assert minecraft_launcher_lib.microsoft_account.url_contains_auth_code("https://login.live.com/oauth20_desktop.srf?code=testcode&lc=test") is True
    assert minecraft_launcher_lib.microsoft_account.url_contains_auth_code("https://example.com") is False


def test_get_auth_code_from_url() -> None:
    assert minecraft_launcher_lib.microsoft_account.get_auth_code_from_url("https://login.live.com/oauth20_desktop.srf?code=testcode&lc=test") == "testcode"
    assert minecraft_launcher_lib.microsoft_account.get_auth_code_from_url("https://login.live.com/oauth20_desktop.srf") is None


def test_parse_auth_code_url() -> None:
    test_state = "123456"
    test_code = "abcxyz"

    auth_code = minecraft_launcher_lib.microsoft_account.parse_auth_code_url(f"http://127.0.0.1:3003?code={test_code}&state={test_state}", test_state)
    assert auth_code == test_code


def test_get_authorization_token(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("default")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    assert minecraft_launcher_lib.microsoft_account.get_authorization_token("client_id", "secret", "redirect_url", "auth_code", "code-verifier")["access_token"] == "test_access_token"


def test_get_store_information(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("default")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    store = minecraft_launcher_lib.microsoft_account.get_store_information("minecraft_access_token")

    assert store["signature"] == "jwt sig"
    assert store["keyId"] == "1"


def test_complete_login(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("default")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    login_data = minecraft_launcher_lib.microsoft_account.complete_login("client_id", None, "redirect_url", "auth_code")

    assert login_data["id"] == "minecraft_uuid"
    assert login_data["name"] == "minecraft_name"
    assert login_data["access_token"] == "minecraft_access_token"
    assert login_data["refresh_token"] == "refresh_token"


def test_complete_login_not_own_minecraft(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("not_own_minecraft")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    with pytest.raises(minecraft_launcher_lib.exceptions.AccountNotOwnMinecraft):
        minecraft_launcher_lib.microsoft_account.complete_login("client_id", None, "redirect_url", "auth_code")


def test_complete_login_not_permitted_azure_app(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("not_permitted_app")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    with pytest.raises(minecraft_launcher_lib.exceptions.AzureAppNotPermitted):
        minecraft_launcher_lib.microsoft_account.complete_login("client_id", None, "redirect_url", "auth_code")


def test_complete_refresh(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("default")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    login_data = minecraft_launcher_lib.microsoft_account.complete_refresh("client_id", "client_secret", None, "refresh_token")

    assert login_data["id"] == "minecraft_uuid"
    assert login_data["name"] == "minecraft_name"
    assert login_data["access_token"] == "minecraft_access_token"
    assert login_data["refresh_token"] == "refresh_token"


def test_complete_refresh_not_own_minecraft(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("not_own_minecraft")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    with pytest.raises(minecraft_launcher_lib.exceptions.AccountNotOwnMinecraft):
        minecraft_launcher_lib.microsoft_account.complete_refresh("client_id", "client_secret", None, "refresh_token")


def test_complete_refresh_invalid_token(monkeypatch: pytest.MonkeyPatch) -> None:
    requests_mock = MicrosoftRequestsMock("default")

    monkeypatch.setattr(requests, "get", requests_mock.get)
    monkeypatch.setattr(requests, "post", requests_mock.post)

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidRefreshToken):
        minecraft_launcher_lib.microsoft_account.complete_refresh("client_id", "client_secret", None, "invalid_token")



from ._test_helper import prepare_test_versions, get_test_callbacks, prepare_requests_mock
from typing import List, Dict, Any
import minecraft_launcher_lib
import pytest_subtests
import requests_mock
import requests
import hashlib
import pathlib
import zipfile
import random
import pytest
import json
import copy
import os


def _create_test_index_pack(index: Dict[str, Any], tmp_path: pathlib.Path, overrides: List[str] = [], client_overrides: List[str] = []) -> str:
    while True:
        name = f"temp-pack-{random.randrange(100, 10000)}.mrpack"
        if not (tmp_path / name).exists():
            temp_pack = str(tmp_path / name)
            break

    with zipfile.ZipFile(temp_pack, "w") as zf:
        zf.writestr("modrinth.index.json", json.dumps(index))

        for i in overrides:
            zf.writestr(f"overrides/{i}", "This is a test override")

        for i in client_overrides:
            zf.writestr(f"client-overrides/{i}", "This is a test client-override")

    return temp_pack


def _create_test_dir(tmp_path: pathlib.Path) -> str:
    while True:
        name = f"temp-dir-{random.randrange(100, 10000)}"
        if not (tmp_path / name).exists():
            return str(tmp_path / name)


_test_file_cache: Dict[str, Dict[str, Any]] = {}


def _generate_test_file(url: str) -> Dict[str, Any]:
    if url in _test_file_cache:
        return _test_file_cache[url]

    r = requests.get(url, stream=True)

    data = {
        "hashes": {
            "sha1": hashlib.sha1(r.content).hexdigest(),
            "sha512": hashlib.sha512(r.content).hexdigest()
        },
        "downloads": [url],
        "fileSize": len(r.content)
    }

    _test_file_cache[url] = data
    return data


def _get_test_index() -> Dict[str, Any]:
    return {
        "formatVersion": 1,
        "name": "Test",
        "versionId": "minecraft-launcher-lib.test",
        "files": [
            {
                **{
                    "path": "a.txt"
                },
                **_generate_test_file("minecraft-launcher-lib-test://text.txt")
            },
            {
                **{
                    "path": "b.txt",
                    "env": {
                        "client": "required",
                        "server": "required"
                    }
                },
                **_generate_test_file("minecraft-launcher-lib-test://text.txt")
            },
            {
                **{
                    "path": "c.txt",
                    "env": {
                        "client": "optional",
                        "server": "optional"
                    }
                },
                **_generate_test_file("minecraft-launcher-lib-test://text.txt")
            },
            {
                **{
                    "path": "d.txt",
                    "env": {
                        "client": "unsupported",
                        "server": "unsupported"
                    }
                },
                **_generate_test_file("minecraft-launcher-lib-test://text.txt")
            },
        ],
        "dependencies": {
            "minecraft": "test1"
        }
    }


def test_get_mrpack_information(requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    prepare_requests_mock(requests_mock)

    index = _get_test_index()

    # Test without summary
    info = minecraft_launcher_lib.mrpack.get_mrpack_information(_create_test_index_pack(index, tmp_path))

    assert info["name"] == "Test"
    assert info["summary"] == ""
    assert info["versionId"] == "minecraft-launcher-lib.test"
    assert info["formatVersion"] == 1
    assert info["minecraftVersion"] == "test1"
    assert info["optionalFiles"] == ["c.txt"]

    # Test with summary
    index["summary"] = "Summary"
    info_summary = minecraft_launcher_lib.mrpack.get_mrpack_information(_create_test_index_pack(index, tmp_path))

    assert info_summary["summary"] == "Summary"


def test_install_mrpack(subtests: pytest_subtests.SubTests, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    prepare_requests_mock(requests_mock)

    index = _get_test_index()

    with subtests.test("Without optional file"):
        first_dir = _create_test_dir(tmp_path)
        minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(index, tmp_path), first_dir, mrpack_install_options={"skipDependenciesInstall": True})
        assert sorted(os.listdir(first_dir)) == sorted(["a.txt", "b.txt"])

    with subtests.test("With optional file"):
        second_dir = _create_test_dir(tmp_path)
        minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(index, tmp_path), second_dir, mrpack_install_options={"skipDependenciesInstall": True, "optionalFiles": ["c.txt"]})
        assert sorted(os.listdir(second_dir)) == sorted(["a.txt", "b.txt", "c.txt"])

    with subtests.test("Overrides"):
        third_dir = _create_test_dir(tmp_path)
        minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(index, tmp_path, overrides=["overrides.txt"], client_overrides=["client/test.txt"]), third_dir, mrpack_install_options={"skipDependenciesInstall": True})
        assert sorted(os.listdir(third_dir)) == sorted(["a.txt", "b.txt", "overrides.txt", "client"])
        assert sorted(os.listdir(os.path.join(third_dir, "client"))) == sorted(["test.txt"])

    with subtests.test("FileOutsideMinecraftDirectory Exception with files"):
        fourth_dir = _create_test_dir(tmp_path)
        exception_index = copy.deepcopy(index)
        exception_index["files"].append({**{
            "path": "../error.txt"
        },
            **_generate_test_file("minecraft-launcher-lib-test://text.txt")
        })
        with pytest.raises(minecraft_launcher_lib.exceptions.FileOutsideMinecraftDirectory):
            minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(exception_index, tmp_path), fourth_dir, mrpack_install_options={"skipDependenciesInstall": True})

    with subtests.test("FileOutsideMinecraftDirectory Exception with overrides"):
        fifth_dir = _create_test_dir(tmp_path)
        with pytest.raises(minecraft_launcher_lib.exceptions.FileOutsideMinecraftDirectory):
            minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(index, tmp_path, overrides=["../overrides.txt"]), fifth_dir, mrpack_install_options={"skipDependenciesInstall": True})

    with subtests.test("Modpack directory"):
        sixth_dir = _create_test_dir(tmp_path)
        modpack_dir = _create_test_dir(tmp_path)
        minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(index, tmp_path), sixth_dir, modpack_directory=modpack_dir, mrpack_install_options={"skipDependenciesInstall": True})
        assert sorted(os.listdir(modpack_dir)) == sorted(["a.txt", "b.txt"])
        assert not os.path.isdir(sixth_dir)

    with subtests.test("Forge"):
        forge_dir = _create_test_dir(tmp_path)
        forge_index = copy.deepcopy(index)
        forge_index["dependencies"]["forge"] = "invalid"
        prepare_test_versions(forge_dir)
        requests_mock.get("https://files.minecraftforge.net/maven/net/minecraftforge/forge/test1-invalid/forge-test1-invalid-installer.jar", status_code=404)
        with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
            minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(forge_index, tmp_path), forge_dir, callback=get_test_callbacks())

    with subtests.test("Fabric"):
        fabric_dir = _create_test_dir(tmp_path)
        fabric_index = copy.deepcopy(index)
        fabric_index["dependencies"]["fabric-loader"] = "invalid"
        prepare_test_versions(fabric_dir)
        with pytest.raises(minecraft_launcher_lib.exceptions.UnsupportedVersion):
            minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(fabric_index, tmp_path), fabric_dir, callback=get_test_callbacks())

    with subtests.test("Quilt"):
        quilt_dir = _create_test_dir(tmp_path)
        quilt_index = copy.deepcopy(index)
        quilt_index["dependencies"]["quilt-loader"] = "invalid"
        prepare_test_versions(quilt_dir)
        with pytest.raises(minecraft_launcher_lib.exceptions.UnsupportedVersion):
            minecraft_launcher_lib.mrpack.install_mrpack(_create_test_index_pack(quilt_index, tmp_path), quilt_dir, callback=get_test_callbacks())


def test_mrpack_launch_version(subtests: pytest_subtests.SubTests, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    prepare_requests_mock(requests_mock)

    index = _get_test_index()

    with subtests.test("Vanilla"):
        assert minecraft_launcher_lib.mrpack.get_mrpack_launch_version(_create_test_index_pack(index, tmp_path)) == "test1"

    with subtests.test("Forge"):
        index["dependencies"]["forge"] = "41.1.0"
        assert minecraft_launcher_lib.mrpack.get_mrpack_launch_version(_create_test_index_pack(index, tmp_path)) == "test1-forge-41.1.0"
        del index["dependencies"]["forge"]

    with subtests.test("Fabric"):
        index["dependencies"]["fabric-loader"] = "0.14.15"
        assert minecraft_launcher_lib.mrpack.get_mrpack_launch_version(_create_test_index_pack(index, tmp_path)) == "fabric-loader-0.14.15-test1"
        del index["dependencies"]["fabric-loader"]

    with subtests.test("Quilt"):
        index["dependencies"]["quilt-loader"] = "0.18.2"
        assert minecraft_launcher_lib.mrpack.get_mrpack_launch_version(_create_test_index_pack(index, tmp_path)) == "quilt-loader-0.18.2-test1"
        del index["dependencies"]["quilt-loader"]



import minecraft_launcher_lib
import platform
import datetime
import pathlib
import pytest
import shutil
import json
import os


def _create_test_version_files(minecraft_directory: pathlib.Path) -> None:
    os.makedirs(os.path.join(minecraft_directory, "versions", "utilstest"))
    with open(os.path.join(minecraft_directory, "versions", "utilstest", "utilstest.json"), "w", encoding="utf-8") as f:
        data = {}
        data["id"] = "utilstest"
        data["type"] = "release"
        data["releaseTime"] = "1970-01-01T00:00:00"
        data["complianceLevel"] = 1
        json.dump(data, f)

    os.makedirs(os.path.join(minecraft_directory, "versions", "invalid-time"))
    with open(os.path.join(minecraft_directory, "versions", "invalid-time", "invalid-time.json"), "w", encoding="utf-8") as f:
        data = {}
        data["id"] = "invalid-time"
        data["type"] = "release"
        data["releaseTime"] = "invalid"
        data["complianceLevel"] = 1
        json.dump(data, f)

    (minecraft_directory / "versions" / "test.txt").touch()


def test_get_minecraft_directory(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    assert minecraft_launcher_lib.utils.get_minecraft_directory() == os.path.join(os.getenv("APPDATA", os.path.join(pathlib.Path.home(), "AppData", "Roaming")), ".minecraft")

    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    assert minecraft_launcher_lib.utils.get_minecraft_directory() == os.path.join(str(pathlib.Path.home()), "Library", "Application Support", "minecraft")

    monkeypatch.setattr(platform, "system", lambda: "Linux")
    assert minecraft_launcher_lib.utils.get_minecraft_directory() == os.path.join(str(pathlib.Path.home()), ".minecraft")


def test_get_latest_version() -> None:
    latest_version = minecraft_launcher_lib.utils.get_latest_version()
    assert "release" in latest_version
    assert "snapshot" in latest_version


def test_get_version_list() -> None:
    version_list = minecraft_launcher_lib.utils.get_version_list()
    for i in version_list:
        assert isinstance(i["id"], str)
        assert isinstance(i["type"], str)
        assert isinstance(i["releaseTime"], datetime.datetime)
        assert isinstance(i["complianceLevel"], int)


def test_get_installed_versions(tmp_path: pathlib.Path) -> None:
    _create_test_version_files(tmp_path)
    version_list = minecraft_launcher_lib.utils.get_installed_versions(tmp_path)
    version_list = minecraft_launcher_lib.utils.get_installed_versions(str(tmp_path))
    for i in version_list:
        assert isinstance(i["id"], str)
        assert isinstance(i["type"], str)
        assert isinstance(i["releaseTime"], datetime.datetime)
        assert isinstance(i["complianceLevel"], int)
    assert len(minecraft_launcher_lib.utils.get_installed_versions("not_existing_directory")) == 0


def test_get_available_versions(tmp_path: pathlib.Path) -> None:
    _create_test_version_files(tmp_path)
    version_list = minecraft_launcher_lib.utils.get_available_versions(tmp_path)
    version_list = minecraft_launcher_lib.utils.get_available_versions(str(tmp_path))
    for i in version_list:
        assert isinstance(i["id"], str)
        assert isinstance(i["type"], str)
        assert isinstance(i["releaseTime"], datetime.datetime)
        assert isinstance(i["complianceLevel"], int)
    assert isinstance(minecraft_launcher_lib.utils.get_available_versions("not_existing_directory"), list)


def test_get_java_executable(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    monkeypatch.setenv("JAVA_HOME", "/testjavahome")
    assert minecraft_launcher_lib.utils.get_java_executable() == os.path.join("/testjavahome", "bin", "javaw.exe")

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    monkeypatch.setattr(os.path, "isfile", lambda path: path == r"C:\Program Files (x86)\Common Files\Oracle\Java\javapath\javaw.exe")
    assert minecraft_launcher_lib.utils.get_java_executable() == r"C:\Program Files (x86)\Common Files\Oracle\Java\javapath\javaw.exe"

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    monkeypatch.setattr(os.path, "isfile", lambda path: False)
    monkeypatch.setattr(shutil, "which", lambda name: "/testwhich" if name == "javaw" else None)
    assert minecraft_launcher_lib.utils.get_java_executable() == "/testwhich"

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(shutil, "which", lambda name: None)
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    monkeypatch.setattr(os.path, "isfile", lambda path: False)
    assert minecraft_launcher_lib.utils.get_java_executable() == "javaw"

    monkeypatch.undo()
    monkeypatch.setenv("JAVA_HOME", "/testjavahome")
    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    assert minecraft_launcher_lib.utils.get_java_executable() == os.path.join("/testjavahome", "bin", "java")

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    monkeypatch.setattr(os.path, "isfile", lambda path: False)
    monkeypatch.setattr(shutil, "which", lambda name: "/testwhich" if name == "java" else None)
    assert minecraft_launcher_lib.utils.get_java_executable() == "/testwhich"

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(shutil, "which", lambda name: None)
    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    monkeypatch.setattr(os.path, "isfile", lambda path: False)
    assert minecraft_launcher_lib.utils.get_java_executable() == "java"

    monkeypatch.undo()
    monkeypatch.setenv("JAVA_HOME", "/testjavahome")
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    assert minecraft_launcher_lib.utils.get_java_executable() == os.path.join("/testjavahome", "bin", "java")

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(os.path, "islink", lambda path: path == "/etc/alternatives/java", raising=False)
    monkeypatch.setattr(os, "readlink", lambda path: "testetc" if path == "/etc/alternatives/java" else None, raising=False)
    assert minecraft_launcher_lib.utils.get_java_executable() == "testetc"

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(os.path, "islink", lambda path: path == "/usr/lib/jvm/default-runtime", raising=False)
    monkeypatch.setattr(os, "readlink", lambda path: "testjvm" if path == "/usr/lib/jvm/default-runtime" else None, raising=False)
    assert minecraft_launcher_lib.utils.get_java_executable() == os.path.join("/usr", "lib", "jvm", "testjvm", "bin", "java")

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(os.path, "isfile", lambda path: False)
    monkeypatch.setattr(os.path, "islink", lambda path: False)
    monkeypatch.setattr(shutil, "which", lambda name: "/testwhich" if name == "java" else None)
    assert minecraft_launcher_lib.utils.get_java_executable() == "/testwhich"

    monkeypatch.undo()
    monkeypatch.delenv("JAVA_HOME", False)
    monkeypatch.setattr(shutil, "which", lambda name: None)
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(os.path, "isfile", lambda path: False)
    monkeypatch.setattr(os.path, "islink", lambda path: False)
    assert minecraft_launcher_lib.utils.get_java_executable() == "java"


def test_get_library_version() -> None:
    assert isinstance(minecraft_launcher_lib.utils.get_library_version(), str)


def test_generate_test_options() -> None:
    options = minecraft_launcher_lib.utils.generate_test_options()
    assert isinstance(options["username"], str)
    assert isinstance(options["uuid"], str)
    assert isinstance(options["token"], str)


def test_is_version_valid(tmp_path: pathlib.Path) -> None:
    _create_test_version_files(tmp_path)
    assert minecraft_launcher_lib.utils.is_version_valid("1.16", tmp_path) is True
    assert minecraft_launcher_lib.utils.is_version_valid("1.16", str(tmp_path)) is True
    assert minecraft_launcher_lib.utils.is_version_valid("utilstest", tmp_path) is True
    assert minecraft_launcher_lib.utils.is_version_valid("utilstest", str(tmp_path)) is True
    assert minecraft_launcher_lib.utils.is_version_valid("Test123", str(tmp_path)) is False
    assert minecraft_launcher_lib.utils.is_version_valid("Test123", tmp_path) is False


def test_get_minecraft_news() -> None:
    assert len(minecraft_launcher_lib.utils.get_minecraft_news()["article_grid"]) == 20
    assert len(minecraft_launcher_lib.utils.get_minecraft_news(page_size=50)["article_grid"]) == 50


def test_is_vanilla_version() -> None:
    assert minecraft_launcher_lib.utils.is_vanilla_version("1.18") is True
    assert minecraft_launcher_lib.utils.is_vanilla_version("test") is False


def test_is_platform_supported(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    assert minecraft_launcher_lib.utils.is_platform_supported() is True

    monkeypatch.setattr(platform, "system", lambda: "Darwin")
    assert minecraft_launcher_lib.utils.is_platform_supported() is True

    monkeypatch.setattr(platform, "system", lambda: "Linux")
    assert minecraft_launcher_lib.utils.is_platform_supported() is True

    monkeypatch.setattr(platform, "system", lambda: "Haiku")
    assert minecraft_launcher_lib.utils.is_platform_supported() is False


def test_is_minecraft_installed(tmp_path: pathlib.Path) -> None:
    assert minecraft_launcher_lib.utils.is_minecraft_installed(tmp_path) is False

    (tmp_path / "versions").mkdir()
    (tmp_path / "libraries").mkdir()
    (tmp_path / "assets").mkdir()

    assert minecraft_launcher_lib.utils.is_minecraft_installed(tmp_path) is True



"""
.. note::
    Before using this module, please read this comment from the forge developers:

    .. code:: text

        Please do not automate the download and installation of Forge.
        Our efforts are supported by ads from the download page.
        If you MUST automate this, please consider supporting the project through https://www.patreon.com/LexManos/

    It's your choice, if you want to respect that and support forge.

forge contains functions for dealing with the Forge modloader
"""
from ._helper import download_file, get_library_path, get_jar_mainclass, parse_maven_metadata, empty, extract_file_from_zip, get_classpath_separator
from .install import install_minecraft_version, install_libraries
from ._internal_types.forge_types import ForgeInstallProfile
from typing import List, Union, Optional
from .exceptions import VersionNotFound
from .types import CallbackDict
import subprocess
import tempfile
import random
import zipfile
import json
import os

__all__ = ["install_forge_version", "run_forge_installer", "list_forge_versions", "find_forge_version", "is_forge_version_valid", "supports_automatic_install", "forge_to_installed_version"]


def get_data_library_path(libname: str, path: str) -> str:
    """
    Turns the libname into a path
    """
    # Remove the []
    libname = libname[1:-1]
    libpath = os.path.join(path, "libraries")
    base_path, libname, version, extra = libname.split(":")
    for i in base_path.split("."):
        libpath = os.path.join(libpath, i)
    try:
        extra, fileend = extra.split("@")
    except ValueError:
        fileend = "jar"
    libpath = os.path.join(libpath, libname, version, f"{libname}-{version}-{extra}.{fileend}")
    return libpath


def forge_processors(data: ForgeInstallProfile, minecraft_directory: Union[str, os.PathLike], lzma_path: str, installer_path: str, callback: CallbackDict, java: str) -> None:
    """
    Run the processors of the install_profile.json
    """
    path = str(minecraft_directory)

    argument_vars = {"{MINECRAFT_JAR}": os.path.join(path, "versions", data["minecraft"], data["minecraft"] + ".jar")}
    for data_key, data_value in data["data"].items():
        if data_value["client"].startswith("[") and data_value["client"].endswith("]"):
            argument_vars["{" + data_key + "}"] = get_data_library_path(data_value["client"], path)
        else:
            argument_vars["{" + data_key + "}"] = data_value["client"]

    with tempfile.TemporaryDirectory() as root_path:
        argument_vars["{INSTALLER}"] = installer_path
        argument_vars["{BINPATCH}"] = lzma_path
        argument_vars["{ROOT}"] = root_path
        argument_vars["{SIDE}"] = "client"

        classpath_seperator = get_classpath_separator()

        callback.get("setMax", empty)(len(data["processors"]))

        for count, i in enumerate(data["processors"]):
            if "client" not in i.get("sides", ["client"]):
                # Skip server side only processors
                continue
            callback.get("setStatus", empty)("Running processor " + i["jar"])
            # Get the classpath
            classpath = ""
            for c in i["classpath"]:
                classpath = classpath + get_library_path(c, path) + classpath_seperator
            classpath = classpath + get_library_path(i["jar"], path)
            mainclass = get_jar_mainclass(get_library_path(i["jar"], path))
            command = [java, "-cp", classpath, mainclass]
            for c in i["args"]:
                var = argument_vars.get(c, c)
                if var.startswith("[") and var.endswith("]"):
                    command.append(get_library_path(var[1:-1], path))
                else:
                    command.append(var)
            for argument_key, argument_value in argument_vars.items():
                for pos in range(len(command)):
                    command[pos] = command[pos].replace(argument_key, argument_value)
            subprocess.run(command)
            callback.get("setProgress", empty)(count)


def install_forge_version(versionid: str, path: Union[str, os.PathLike], callback: Optional[CallbackDict] = None, java: Optional[Union[str, os.PathLike]] = None) -> None:
    """
    Installs the given Forge version

    :param versionid: A Forge Version. You can get a List of Forge versions using :func:`list_forge_versions`
    :param path: The path to your Minecraft directory
    :param callback: The same dict as for :func:`~minecraft_launcher_lib.install.install_minecraft_version`
    :param java: A Path to a custom Java executable

    Raises a :class:`~minecraft_launcher_lib.exceptions.VersionNotFound` exception when the given forge version is not found
    """
    if callback is None:
        callback = {}

    FORGE_DOWNLOAD_URL = "https://files.minecraftforge.net/maven/net/minecraftforge/forge/{version}/forge-{version}-installer.jar"
    temp_file_path = os.path.join(tempfile.gettempdir(), "forge-installer-" + str(random.randrange(1, 100000)) + ".tmp")

    if not download_file(FORGE_DOWNLOAD_URL.format(version=versionid), temp_file_path, callback):
        raise VersionNotFound(versionid)

    zf = zipfile.ZipFile(temp_file_path, "r")

    # Read the install_profile.json
    with zf.open("install_profile.json", "r") as f:
        version_content = f.read()

    version_data: ForgeInstallProfile = json.loads(version_content)
    forge_version_id = version_data["version"] if "version" in version_data else version_data["install"]["version"]
    minecraft_version = version_data["minecraft"] if "minecraft" in version_data else version_data["install"]["minecraft"]

    # Make sure, the base version is installed
    install_minecraft_version(minecraft_version, path, callback=callback)

    # Install all needed libs from install_profile.json
    if "libraries" in version_data:
        install_libraries(minecraft_version, version_data["libraries"], str(path), callback)

    # Extract the version.json
    version_json_path = os.path.join(path, "versions", forge_version_id, forge_version_id + ".json")
    try:
        extract_file_from_zip(zf, "version.json", version_json_path, minecraft_directory=path)
    except KeyError:
        if "versionInfo" in version_data:
            with open(version_json_path, "w", encoding="utf-8") as f:
                json.dump(version_data["versionInfo"], f, ensure_ascii=False, indent=4)

    # Extract forge libs from the installer
    forge_lib_path = os.path.join(path, "libraries", "net", "minecraftforge", "forge", versionid)
    try:
        extract_file_from_zip(zf, "maven/net/minecraftforge/forge/{version}/forge-{version}-universal.jar".format(version=versionid), os.path.join(forge_lib_path, "forge-" + versionid + "-universal.jar"), minecraft_directory=path)
    except KeyError:
        pass

    try:
        extract_file_from_zip(zf, "forge-{version}-universal.jar".format(version=versionid), os.path.join(forge_lib_path, f"forge-{versionid}.jar"), minecraft_directory=path)
    except KeyError:
        pass

    try:
        extract_file_from_zip(zf, f"maven/net/minecraftforge/forge/{versionid}/forge-{versionid}.jar", os.path.join(forge_lib_path, f"forge-{versionid}.jar"), minecraft_directory=path)
    except KeyError:
        pass

    # Extract the client.lzma
    lzma_path = os.path.join(tempfile.gettempdir(), "lzma-" + str(random.randrange(1, 100000)) + ".tmp")
    try:
        extract_file_from_zip(zf, "data/client.lzma", lzma_path)
    except KeyError:
        pass

    zf.close()

    # Install the rest with the vanilla function
    install_minecraft_version(forge_version_id, str(path), callback=callback)

    # Run the processors
    if "processors" in version_data:
        forge_processors(version_data, str(path), lzma_path, temp_file_path, callback, "java" if java is None else str(java))

    # Delete the temporary files
    os.remove(temp_file_path)
    if os.path.isfile(lzma_path):
        os.remove(lzma_path)


def run_forge_installer(version: str, java: Optional[Union[str, os.PathLike]] = None) -> None:
    """
    Run the forge installer of the given forge version

    :param version: A Forge Version. You can get a List of Forge versions using :func:`list_forge_versions`
    :param java: A Path to a custom Java executable
    """
    FORGE_DOWNLOAD_URL = "https://files.minecraftforge.net/maven/net/minecraftforge/forge/{version}/forge-{version}-installer.jar"
    temp_file_path = os.path.join(tempfile.gettempdir(), "forge-" + str(random.randrange(1, 100000)) + ".tmp")

    if not download_file(FORGE_DOWNLOAD_URL.format(version=version), temp_file_path, {}):
        raise VersionNotFound(version)

    try:
        subprocess.run(["java" if java is None else str(java), "-jar", temp_file_path], check=True)
    finally:
        os.remove(temp_file_path)


def list_forge_versions() -> List[str]:
    """
    Returns a list of all forge versions
    """
    MAVEN_METADATA_URL = "https://files.minecraftforge.net/maven/net/minecraftforge/forge/maven-metadata.xml"
    return parse_maven_metadata(MAVEN_METADATA_URL)["versions"]


def find_forge_version(vanilla_version: str) -> Optional[str]:
    """
    Find the latest forge version that is compatible to the given vanilla version

    :param vanilla_version: A vanilla Minecraft version
    """
    version_list = list_forge_versions()
    for i in version_list:
        version_split = i.split("-")
        if version_split[0] == vanilla_version:
            return i
    return None


def is_forge_version_valid(forge_version: str) -> bool:
    """
    Checks if a forge version is valid

    :param forge_version: A Forge Version
    """
    forge_version_list = list_forge_versions()
    return forge_version in forge_version_list


def supports_automatic_install(forge_version: str) -> bool:
    """
    Checks if install_forge_version() supports the given forge version

    :param forge_version: A Forge Version
    """
    try:
        vanilla_version, forge = forge_version.split("-")
        version_split = vanilla_version.split(".")
        version_number = int(version_split[1])
        if version_number >= 13:
            return True
        else:
            return False
    except Exception:
        return False


def forge_to_installed_version(forge_version: str) -> str:
    """
    Returns the Version under which Forge will be installed from the given Forge version.

    :param forge_version: A Forge Version

    Raises a ValueError if the Version is invalid.
    """
    try:
        vanilla_part, forge_part = forge_version.split("-")
        return f"{vanilla_part}-forge-{forge_part}"
    except ValueError:
        raise ValueError(f"{forge_version} is not a valid forge version") from None



"""
This module contains all Types for minecraft-launcher-lib. It may help your IDE. You don't need to use this module directly in your code.
If you are not interested in static typing just ignore it.
For more information about TypeDict see `PEP 589 <https://peps.python.org/pep-0589/>`_.
"""
from typing import Literal, TypedDict, List, Callable, Optional
import datetime


class MinecraftOptions(TypedDict, total=False):
    username: str
    uuid: str
    token: str
    executablePath: str
    defaultExecutablePath: str
    jvmArguments: List[str]
    launcherName: str
    launcherVersion: str
    gameDirectory: str
    demo: bool
    customResolution: bool
    resolutionWidth: str
    resolutionHeight: str
    server: str
    port: str
    nativesDirectory: str
    enableLoggingConfig: bool
    disableMultiplayer: bool
    disableChat: bool
    quickPlayPath: Optional[str]
    quickPlaySingleplayer: Optional[str]
    quickPlayMultiplayer: Optional[str]
    quickPlayRealms: Optional[str]


class CallbackDict(TypedDict, total=False):
    setStatus: Callable[[str], None]
    setProgress: Callable[[int], None]
    setMax: Callable[[int], None]


class LatestMinecraftVersions(TypedDict):
    release: str
    snapshot: str


class MinecraftVersionInfo(TypedDict):
    id: str
    type: str
    releaseTime: datetime.datetime
    complianceLevel: int


# fabric

class FabricMinecraftVersion(TypedDict):
    version: str
    stable: bool


class FabricLoader(TypedDict):
    separator: str
    build: int
    maven: str
    version: str
    stable: bool


# quilt

class QuiltMinecraftVersion(TypedDict):
    version: str
    stable: bool


class QuiltLoader(TypedDict):
    separator: str
    build: int
    maven: str
    version: str

# ----
# News
# ----


class _ImageBase(TypedDict):
    content_type: Literal["image", "outgoing-link", "video"]
    imageURL: str


class Image(_ImageBase, total=False):
    alt: str
    videoURL: str
    videoType: str
    videoProvider: str
    videoId: str
    linkurl: str
    background_color: Literal["bg-blue", "bg-green", "bg-red"]


class Tile(TypedDict):
    sub_header: str
    image: Image
    tile_size: Literal["1x1", "1x2", "2x1", "2x2", "4x2"]
    title: str


class _ArticleBase(TypedDict):
    default_tile: Tile
    articleLang: Literal["en-us"]
    primary_category: str
    categories: List[str]
    article_url: str
    publish_date: str
    tags: List[str]


class Article(_ArticleBase, total=False):
    preferred_tile: Tile


class Articles(TypedDict):
    article_grid: List[Article]
    article_count: int


# java_utils

class JavaInformation(TypedDict):
    path: str
    name: str
    version: str
    java_path: str
    javaw_path: Optional[str]
    is_64bit: bool
    openjdk: bool


# vanilla_launcher

class VanillaLauncherProfileResolution(TypedDict):
    height: int
    width: int


class VanillaLauncherProfile(TypedDict, total=False):
    name: str
    version: Optional[str]
    versionType: Literal["latest-release", "latest-snapshot", "custom"]
    gameDirectory: Optional[str]
    javaExecutable: Optional[str]
    javaArguments: Optional[List[str]]
    customResolution: Optional[VanillaLauncherProfileResolution]


# mrpack

class MrpackInformation(TypedDict):
    name: str
    summary: str
    versionId: str
    formatVersion: int
    minecraftVersion: str
    optionalFiles: List[str]


class MrpackInstallOptions(TypedDict, total=False):
    optionalFiles: List[str]
    skipDependenciesInstall: bool


# runtime

class JvmRuntimeInformation(TypedDict):
    name: str
    released: datetime.datetime



"vanilla_launcher contains some functions for interacting with the Vanilla Minecraft Launcher"
from ._internal_types.vanilla_launcher_types import VanillaLauncherProfilesJson, VanillaLauncherProfilesJsonProfile
from .types import VanillaLauncherProfile, MinecraftOptions
from .exceptions import InvalidVanillaLauncherProfile
from .utils import get_latest_version
from ._helper import assert_func
from typing import List, Union
import datetime
import json
import uuid
import os

__all__ = [
    "load_vanilla_launcher_profiles",
    "vanilla_launcher_profile_to_minecraft_options",
    "get_vanilla_launcher_profile_version",
    "add_vanilla_launcher_profile",
    "do_vanilla_launcher_profiles_exists"
]


def _is_vanilla_launcher_profile_valid(vanilla_profile: VanillaLauncherProfile) -> bool:
    "Checks if the given profile is valid"
    if not isinstance(vanilla_profile.get("name"), str):
        return False

    if vanilla_profile.get("versionType") not in ("latest-release", "latest-snapshot", "custom"):
        return False

    if vanilla_profile["versionType"] == "custom" and vanilla_profile.get("version") is None:
        return False

    if vanilla_profile.get("gameDirectory") is not None and not isinstance(vanilla_profile.get("gameDirectory"), str):
        return False

    if vanilla_profile.get("javaExecutable") is not None and not isinstance(vanilla_profile.get("javaExecutable"), str):
        return False

    if (java_arguments := vanilla_profile.get("javaArguments")) is not None:
        try:
            for i in java_arguments:
                assert_func(isinstance(i, str))
        except Exception:
            return False

    if (custom_resolution := vanilla_profile.get("customResolution")) is not None:
        try:
            assert_func(len(custom_resolution) == 2)
            assert_func(isinstance(custom_resolution["height"], int))
            assert_func(isinstance(custom_resolution["width"], int))
        except Exception:
            return False

    return True


def load_vanilla_launcher_profiles(minecraft_directory: Union[str, os.PathLike]) -> List[VanillaLauncherProfile]:
    """
    Loads the profiles of the Vanilla Launcher from the given Minecraft directory

    :param minecraft_directory: The Minecraft directory
    :return: A List with the Profiles
    """
    with open(os.path.join(minecraft_directory, "launcher_profiles.json"), "r", encoding="utf-8") as f:
        data: VanillaLauncherProfilesJson = json.load(f)

    profile_list: List[VanillaLauncherProfile] = []
    for value in data["profiles"].values():
        vanilla_profile: VanillaLauncherProfile = {}

        if value["type"] == "latest-release":
            vanilla_profile["name"] = "Latest release"
        elif value["type"] == "latest-snapshot":
            vanilla_profile["name"] = "Latest snapshot"
        else:
            vanilla_profile["name"] = value["name"]

        if value["lastVersionId"] == "latest-release":
            vanilla_profile["versionType"] = "latest-release"
            vanilla_profile["version"] = None
        elif value["lastVersionId"] == "latest-snapshot":
            vanilla_profile["versionType"] = "latest-snapshot"
            vanilla_profile["version"] = None
        else:
            vanilla_profile["versionType"] = "custom"
            vanilla_profile["version"] = value["lastVersionId"]

        vanilla_profile["gameDirectory"] = value.get("gameDir")
        vanilla_profile["javaExecutable"] = value.get("javaDir")

        if "javaArgs" in value:
            vanilla_profile["javaArguments"] = value["javaArgs"].split(" ")
        else:
            vanilla_profile["javaArguments"] = None

        if "resolution" in value:
            vanilla_profile["customResolution"] = {
                "height": value["resolution"]["height"],
                "width": value["resolution"]["width"]
            }
        else:
            vanilla_profile["customResolution"] = None

        profile_list.append(vanilla_profile)

    return profile_list


def vanilla_launcher_profile_to_minecraft_options(vanilla_profile: VanillaLauncherProfile) -> MinecraftOptions:
    """
    Converts a VanillaLauncherProfile into a Options dict, that can be used by :func:`~minecraft_launcher_lib.install.install_minecraft_version`.
    You still need to add the Login Data to the Options before you can use it.

    :param vanilla_profile: The profile as returned by :func:`load_vanilla_launcher_profiles`
    :raises InvalidVanillaLauncherProfile: The given Profile is invalid
    :return: The Options Dict
    """
    if not _is_vanilla_launcher_profile_valid(vanilla_profile):
        raise InvalidVanillaLauncherProfile(vanilla_profile)

    options: MinecraftOptions = {}

    if (game_directory := vanilla_profile.get("gameDirectory")) is not None:
        options["gameDirectory"] = game_directory

    if (java_executable := vanilla_profile.get("javaExecutable")) is not None:
        options["executablePath"] = java_executable

    if (java_arguments := vanilla_profile.get("javaArguments")) is not None:
        options["jvmArguments"] = java_arguments

    if (custom_resolution := vanilla_profile.get("customResolution")) is not None:
        options["customResolution"] = True
        options["resolutionWidth"] = str(custom_resolution["width"])
        options["resolutionHeight"] = str(custom_resolution["height"])

    return options


def get_vanilla_launcher_profile_version(vanilla_profile: VanillaLauncherProfile) -> str:
    """
    Returns the Minecraft version of the VanillaProfile. Handles ``latest-release`` and ``latest-snapshot``.

    :param vanilla_profile: The Profile
    :type vanilla_profile: VanillaLauncherProfile
    :raises InvalidVanillaLauncherProfile: The given Profile is invalid
    :return: The Minecraft version
    """
    if not _is_vanilla_launcher_profile_valid(vanilla_profile):
        raise InvalidVanillaLauncherProfile(vanilla_profile)

    if vanilla_profile["versionType"] == "latest-release":
        return get_latest_version()["release"]
    elif vanilla_profile["versionType"] == "latest-snapshot":
        return get_latest_version()["snapshot"]
    elif vanilla_profile["versionType"] == "custom":
        return vanilla_profile["version"]  # type: ignore


def add_vanilla_launcher_profile(minecraft_directory: Union[str, os.PathLike], vanilla_profile: VanillaLauncherProfile) -> None:
    """
    Adds a new Profile to the Vanilla Launcher

    :param minecraft_directory: The Minecraft directory
    :param vanilla_profile: The new Profile
    :raises InvalidVanillaLauncherProfile: The given Profile is invalid
    """
    if not _is_vanilla_launcher_profile_valid(vanilla_profile):
        raise InvalidVanillaLauncherProfile(vanilla_profile)

    with open(os.path.join(minecraft_directory, "launcher_profiles.json"), "r", encoding="utf-8") as f:
        data: VanillaLauncherProfilesJson = json.load(f)

    new_profile: VanillaLauncherProfilesJsonProfile = {}
    new_profile["name"] = vanilla_profile["name"]

    if vanilla_profile["versionType"] == "latest-release":
        new_profile["lastVersionId"] = "latest-release"
    elif vanilla_profile["versionType"] == "latest-snapshot":
        new_profile["lastVersionId"] = "latest-snapshot"
    elif vanilla_profile["versionType"] == "custom":
        # _is_vanilla_launcher_profile_valid() ensures that version is not None, when versionType is set to custom, so we can ignore the mypy error here
        new_profile["lastVersionId"] = vanilla_profile["version"]  # type: ignore

    if (game_directory := vanilla_profile.get("gameDirectory")) is not None:
        new_profile["gameDir"] = game_directory

    if (java_executable := vanilla_profile.get("javaExecutable")) is not None:
        new_profile["javaDir"] = java_executable

    if (java_arguments := vanilla_profile.get("javaArguments")) is not None:
        new_profile["javaArgs"] = " ".join(java_arguments)

    if (custom_resolution := vanilla_profile.get("customResolution")) is not None:
        new_profile["resolution"] = {
            "height": custom_resolution["height"],
            "width": custom_resolution["width"]
        }

    new_profile["created"] = datetime.datetime.now().isoformat()
    new_profile["lastUsed"] = datetime.datetime.now().isoformat()
    new_profile["type"] = "custom"

    # Generate a Key for the Profile
    while True:
        key = str(uuid.uuid4())
        if key not in data["profiles"]:
            break

    data["profiles"][key] = new_profile

    with open(os.path.join(minecraft_directory, "launcher_profiles.json"), "w", encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def do_vanilla_launcher_profiles_exists(minecraft_directory: Union[str, os.PathLike]) -> bool:
    """
    Checks if profiles from the vanilla launcher can be found

    :param minecraft_directory: The Minecraft directory
    :return: If profiles exists
    """
    return os.path.isfile(os.path.join(minecraft_directory, "launcher_profiles.json"))



"install allows you to install minecraft."
from ._helper import download_file, parse_rule_list, inherit_json, empty, get_user_agent, check_path_inside_minecraft_directory
from ._internal_types.shared_types import ClientJson, ClientJsonLibrary
from .natives import extract_natives_file, get_natives
from ._internal_types.install_types import AssetsJson
from .runtime import install_jvm_runtime
from typing import List, Optional, Union
from .exceptions import VersionNotFound
from .types import CallbackDict
import requests
import shutil
import json
import os

__all__ = ["install_minecraft_version"]


def install_libraries(id: str, libraries: List[ClientJsonLibrary], path: str, callback: CallbackDict) -> None:
    """
    Install all libraries
    """
    session = requests.session()
    callback.get("setStatus", empty)("Download Libraries")
    callback.get("setMax", empty)(len(libraries) - 1)
    for count, i in enumerate(libraries):
        # Check, if the rules allow this lib for the current system
        if "rules" in i and not parse_rule_list(i["rules"], {}):
            continue

        # Turn the name into a path
        current_path = os.path.join(path, "libraries")
        if "url" in i:
            if i["url"].endswith("/"):
                download_url = i["url"][:-1]
            else:
                download_url = i["url"]
        else:
            download_url = "https://libraries.minecraft.net"

        try:
            lib_path, name, version = i["name"].split(":")[0:3]
        except ValueError:
            continue

        for lib_part in lib_path.split("."):
            current_path = os.path.join(current_path, lib_part)
            download_url = f"{download_url}/{lib_part}"

        try:
            version, fileend = version.split("@")
        except ValueError:
            fileend = "jar"

        jar_filename = f"{name}-{version}.{fileend}"
        download_url = f"{download_url}/{name}/{version}"
        current_path = os.path.join(current_path, name, version)
        native = get_natives(i)

        # Check if there is a native file
        if native != "":
            jar_filename_native = f"{name}-{version}-{native}.jar"
        jar_filename = f"{name}-{version}.{fileend}"
        download_url = f"{download_url}/{jar_filename}"

        # Try to download the lib
        try:
            download_file(download_url, os.path.join(current_path, jar_filename), callback=callback, session=session, minecraft_directory=path)
        except Exception:
            pass

        if "downloads" not in i:
            if "extract" in i:
                extract_natives_file(os.path.join(current_path, jar_filename_native), os.path.join(path, "versions", id, "natives"), i["extract"])
            continue

        if "artifact" in i["downloads"] and i["downloads"]["artifact"]["url"] != "" and "path" in i["downloads"]["artifact"]:
            download_file(i["downloads"]["artifact"]["url"], os.path.join(path, "libraries", i["downloads"]["artifact"]["path"]), callback, sha1=i["downloads"]["artifact"]["sha1"], session=session, minecraft_directory=path)
        if native != "":
            download_file(i["downloads"]["classifiers"][native]["url"], os.path.join(current_path, jar_filename_native), callback, sha1=i["downloads"]["classifiers"][native]["sha1"], session=session, minecraft_directory=path)  # type: ignore
            if "extract" in i:
                extract_natives_file(os.path.join(current_path, jar_filename_native), os.path.join(path, "versions", id, "natives"), i["extract"])

        callback.get("setProgress", empty)(count)


def install_assets(data: ClientJson, path: str, callback: CallbackDict) -> None:
    """
    Install all assets
    """
    # Old versions don't have this
    if "assetIndex" not in data:
        return

    callback.get("setStatus", empty)("Download Assets")
    session = requests.session()

    # Download all assets
    download_file(data["assetIndex"]["url"], os.path.join(path, "assets", "indexes", data["assets"] + ".json"), callback, sha1=data["assetIndex"]["sha1"], session=session)
    with open(os.path.join(path, "assets", "indexes", data["assets"] + ".json")) as f:
        assets_data: AssetsJson = json.load(f)

    # The assets has a hash. e.g. c4dbabc820f04ba685694c63359429b22e3a62b5
    # With this hash, it can be download from https://resources.download.minecraft.net/c4/c4dbabc820f04ba685694c63359429b22e3a62b5
    # And saved at assets/objects/c4/c4dbabc820f04ba685694c63359429b22e3a62b5
    callback.get("setMax", empty)(len(assets_data["objects"]) - 1)
    count = 0
    for value in assets_data["objects"].values():
        download_file("https://resources.download.minecraft.net/" + value["hash"][:2] + "/" + value["hash"], os.path.join(path, "assets", "objects", value["hash"][:2], value["hash"]), callback, sha1=value["hash"], session=session, minecraft_directory=path)
        count += 1
        callback.get("setProgress", empty)(count)


def do_version_install(versionid: str, path: str, callback: CallbackDict, url: Optional[str] = None, sha1: Optional[str] = None) -> None:
    """
    Installs the given version
    """
    # Download and read versions.json
    if url:
        download_file(url, os.path.join(path, "versions", versionid, versionid + ".json"), callback, sha1=sha1, minecraft_directory=path)

    with open(os.path.join(path, "versions", versionid, versionid + ".json"), "r", encoding="utf-8") as f:
        versiondata: ClientJson = json.load(f)

    # For Forge
    if "inheritsFrom" in versiondata:
        try:
            install_minecraft_version(versiondata["inheritsFrom"], path, callback=callback)
        except VersionNotFound:
            pass
        versiondata = inherit_json(versiondata, path)

    install_libraries(versiondata["id"], versiondata["libraries"], path, callback)
    install_assets(versiondata, path, callback)

    # Download logging config
    if "logging" in versiondata:
        if len(versiondata["logging"]) != 0:
            logger_file = os.path.join(path, "assets", "log_configs", versiondata["logging"]["client"]["file"]["id"])
            download_file(versiondata["logging"]["client"]["file"]["url"], logger_file, callback, sha1=versiondata["logging"]["client"]["file"]["sha1"], minecraft_directory=path)

    # Download minecraft.jar
    if "downloads" in versiondata:
        download_file(versiondata["downloads"]["client"]["url"], os.path.join(path, "versions", versiondata["id"], versiondata["id"] + ".jar"), callback, sha1=versiondata["downloads"]["client"]["sha1"], minecraft_directory=path)

    # Need to copy jar for old forge versions
    if not os.path.isfile(os.path.join(path, "versions", versiondata["id"], versiondata["id"] + ".jar")) and "inheritsFrom" in versiondata:
        inherits_from = versiondata["inheritsFrom"]
        inherit_path = os.path.join(path, "versions", inherits_from, f"{inherits_from}.jar")
        check_path_inside_minecraft_directory(path, inherit_path)
        shutil.copyfile(os.path.join(path, "versions", versiondata["id"], versiondata["id"] + ".jar"), inherit_path)

    # Install java runtime if needed
    if "javaVersion" in versiondata:
        callback.get("setStatus", empty)("Install java runtime")
        install_jvm_runtime(versiondata["javaVersion"]["component"], path, callback=callback)

    callback.get("setStatus", empty)("Installation complete")


def install_minecraft_version(versionid: str, minecraft_directory: Union[str, os.PathLike], callback: Optional[CallbackDict] = None) -> None:
    """
    Installs a minecraft version into the given path. e.g. ``install_version("1.14", "/tmp/minecraft")``. Use :func:`~minecraft_launcher_lib.utils.get_minecraft_directory` to get the default Minecraft directory.

    :param versionid: The Minecraft version
    :param minecraft_directory: The path to your Minecraft directory
    :param callback: Some functions that are called to monitor the progress (see below)
    :raises VersionNotFound: The Minecraft version was not found
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory

    ``callback`` is a dict with functions that are called with arguments to get the progress. You can use it to show the progress to the user.

    .. code:: python

        callback = {
            "setStatus": some_function, # This function is called to set a text
            "setProgress" some_function, # This function is called to set the progress.
            "setMax": some_function, # This function is called to set to max progress.
        }

    Files that are already exists will not be replaced.
    """
    if isinstance(minecraft_directory, os.PathLike):
        minecraft_directory = str(minecraft_directory)
    if callback is None:
        callback = {}
    if os.path.isfile(os.path.join(minecraft_directory, "versions", versionid, f"{versionid}.json")):
        do_version_install(versionid, minecraft_directory, callback)
        return
    version_list = requests.get("https://launchermeta.mojang.com/mc/game/version_manifest_v2.json", headers={"user-agent": get_user_agent()}).json()
    for i in version_list["versions"]:
        if i["id"] == versionid:
            do_version_install(versionid, minecraft_directory, callback, url=i["url"], sha1=i["sha1"])
            return
    raise VersionNotFound(versionid)



"java_utils contains some functions to help with Java"
from typing import List, Union, Optional
from .types import JavaInformation
import subprocess
import platform
import re
import os


def get_java_information(path: Union[str, os.PathLike]) -> JavaInformation:
    """
    Returns Some Information about the given Java Installation

    :param path: The Path to the Installation. It must be the Directory. If your Java executbale is e.g. /usr/lib/jvm/java-19-openjdk-amd64/bin/java this Parameter must be /usr/lib/jvm/java-19-openjdk-amd64.
    :return: A dict with Information about the given java installation

    Raises a ValueError on the worng Path

    .. note::

        This Function executes the Java executable to detemine details such as the version. This might be a security risk.
    """
    if platform.system() == "Windows":
        if not os.path.isfile(os.path.join(path, "bin", "java.exe")):
            raise ValueError(os.path.abspath(os.path.join(path, "bin", "java.exe")) + " was not found")
    else:
        if not os.path.isfile(os.path.join(path, "bin", "java")):
            raise ValueError(os.path.abspath(os.path.join(path, "bin", "java")) + " was not found")

    lines = subprocess.run([os.path.join(path, "bin", "java"), "-showversion"], capture_output=True, text=True).stderr.splitlines()
    information: JavaInformation = {}  # type: ignore
    information["path"] = str(path)
    information["name"] = os.path.basename(path)
    information["version"] = re.search(r'(?<=version ")[\d|\.|_]+(?=")', lines[0]).group()  # type: ignore
    information["is_64bit"] = "64-Bit" in lines[2]
    information["openjdk"] = lines[0].startswith("openjdk")

    if platform.system() == "Windows":
        information["java_path"] = os.path.join(os.path.abspath(path), "bin", "java.exe")
        information["javaw_path"] = os.path.join(os.path.abspath(path), "bin", "javaw.exe")
    else:
        information["java_path"] = os.path.join(os.path.abspath(path), "bin", "java")
        information["javaw_path"] = None

    return information


def _search_java_directory(path: Union[str, os.PathLike]) -> List[str]:
    "Used by find_system_java_versions() to parse a Directory"
    if not os.path.isdir(path):
        return []

    java_list: List[str] = []
    for i in os.listdir(path):
        current_entry = os.path.join(path, i)

        if os.path.isfile(current_entry) or os.path.islink(current_entry):
            continue

        if os.path.isfile(os.path.join(current_entry, "bin", "java")) or os.path.isfile(os.path.join(current_entry, "bin", "java.exe")):
            java_list.append(current_entry)

    return java_list


def find_system_java_versions(additional_directories: Optional[List[Union[str, os.PathLike]]] = None) -> List[str]:
    """
    Try to find all Java Versions installed on the System. You can use this to e.g. let the User choose between different Java Versions in a Dropdown.

    :param additional_directories: A List of additional Directories to search for Java in custom locations
    :return: A List with all Directories of Java Installations

    macOS is not supported yet
    """
    java_list: List[str] = []

    if platform.system() == "Windows":
        java_list += _search_java_directory(r"C:\Program Files (x86)\Java")
        java_list += _search_java_directory(r"C:\Program Files\Java")
    elif platform.system() == "Linux":
        java_list += _search_java_directory("/usr/lib/jvm")
        java_list += _search_java_directory("/usr/lib/sdk")

    if additional_directories is not None:
        for i in additional_directories:
            java_list += _search_java_directory(i)

    return java_list


def find_system_java_versions_information(additional_directories: Optional[List[Union[str, os.PathLike]]] = None) -> List[JavaInformation]:
    """
    Same as :func:`find_system_java_version`, but uses :func:`get_java_information` to get some Information about the Installation instead of just proving a Path.

    :param additional_directories: A List of additional Directories to search for Java in custom locations
    :return: A List with Information of Java Installations

    macOS is not supported yet

    .. note::

        This Function executes the Java executable to detemine details such as the version. This might be a security risk.
    """
    java_information_list: List[JavaInformation] = []
    for i in find_system_java_versions(additional_directories=additional_directories):
        java_information_list.append(get_java_information(i))
    return java_information_list



from . import command, install, microsoft_account, utils, java_utils, forge, fabric, quilt, runtime, vanilla_launcher, mrpack, exceptions, types, microsoft_types
__all__ = ["command", "install", "microsoft_account", "utils", "java_utils", "forge", "fabric", "quilt", "runtime", "vanilla_launcher", "mrpack", "exceptions", "types", "microsoft_types"]



from .shared_types import ClientJson, ClientJsonLibrary
from typing import List, Dict, Literal, TypedDict


class _ForgeInstallProcessor(TypedDict, total=False):
    sides: List[Literal["client", "server"]]
    jar: str
    classpath: List[str]
    args: List[str]


class _ForgeInstallProfileInstall(TypedDict, total=False):
    profileName: str
    target: str
    path: str
    version: str
    filePath: str
    welcome: str
    minecraft: str
    mirrorList: str
    logo: str


class ForgeInstallProfile(TypedDict, total=False):
    spec: int
    profile: str
    version: str
    minecraft: str
    serverJarPath: str
    data: Dict[str, Dict[Literal["client", "server"], str]]
    processors: List[_ForgeInstallProcessor]
    libraries: List[ClientJsonLibrary]
    icon: str
    logo: str
    mirrorList: str
    welcome: str
    install: _ForgeInstallProfileInstall
    versionInfo: ClientJson



from typing import List, TypedDict, Literal


class MrpackFileHashes(TypedDict):
    sha1: str
    sha256: str


class MrpackFileEnv(TypedDict):
    client: Literal["required", "optional", "unsupported"]
    server: Literal["required", "optional", "unsupported"]


class MrpackFile(TypedDict, total=False):
    path: str
    hashes: MrpackFileHashes
    env: MrpackFileEnv
    downloads: List[str]
    fileSize: int


MrpackDependencies = TypedDict("MrpackDependencies", {
    "minecraft": str,
    "forge": str,
    "fabric-loader": str,
    "quilt-loader": str
}, total=False)


class MrpackIndex(TypedDict, total=False):
    formatVersion: int
    game: str
    versionId: str
    name: str
    summary: str
    files: List[MrpackFile]
    dependencies: MrpackDependencies



from ._test_helper import prepare_test_versions, get_test_callbacks, prepare_requests_mock
import minecraft_launcher_lib
import requests_mock
import subprocess
import platform
import zipfile
import pathlib
import pytest
import io


def _create_bytes_zip(source_dir: pathlib.Path) -> bytes:
    buffer = io.BytesIO()
    zf = zipfile.ZipFile(buffer, "w")
    for current_file in source_dir.rglob("*"):
        if current_file.is_file():
            zf.writestr(str(current_file.relative_to(source_dir)), current_file.read_bytes())
    zf.close()
    return buffer.getvalue()


def test_install_forge_version(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    requests_mock.get("https://files.minecraftforge.net/maven/net/minecraftforge/forge/forgetest1/forge-forgetest1-installer.jar", content=_create_bytes_zip(pathlib.Path(__file__).parent / "data" / "forge" / "forgetest1"))
    requests_mock.get("https://files.minecraftforge.net/maven/net/minecraftforge/forge/forgetest2/forge-forgetest2-installer.jar", content=_create_bytes_zip(pathlib.Path(__file__).parent / "data" / "forge" / "forgetest2"))
    requests_mock.real_http = True

    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))
    monkeypatch.setattr(subprocess, "run", lambda cmd, **kwargs: None)

    prepare_test_versions(tmp_path)
    prepare_requests_mock(requests_mock)

    minecraft_launcher_lib.forge.install_forge_version("forgetest1", tmp_path, callback=get_test_callbacks())
    minecraft_launcher_lib.forge.install_forge_version("forgetest2", tmp_path, callback=get_test_callbacks())

    assert (tmp_path / "libraries" / "net" / "minecraftforge" / "forge" / "forgetest1" / "forge-forgetest1.jar").is_file()


def test_install_forge_version_invalid_version(requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    requests_mock.get("https://files.minecraftforge.net/maven/net/minecraftforge/forge/invalid/forge-invalid-installer.jar", status_code=404)

    # Checks if the VersionNotFound exception raised
    with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound) as ex:
        minecraft_launcher_lib.forge.install_forge_version("invalid", str(tmp_path))

    assert ex.value.version == "invalid"


def test_run_forge_installer(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker) -> None:
    requests_mock.get("https://files.minecraftforge.net/maven/net/minecraftforge/forge/test/forge-test-installer.jar", text="Hello")
    requests_mock.get("https://files.minecraftforge.net/maven/net/minecraftforge/forge/invalid/forge-invalid-installer.jar", text="World", status_code=404)
    monkeypatch.setattr(subprocess, "run", lambda cmd, **kwargs: exec("raise subprocess.CalledProcessError(1, [])") if cmd[0] != "java" else None)

    minecraft_launcher_lib.forge.run_forge_installer("test")

    with pytest.raises(subprocess.CalledProcessError):
        minecraft_launcher_lib.forge.run_forge_installer("test", java="test")

    with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
        minecraft_launcher_lib.forge.run_forge_installer("invalid")


def test_list_forge_versions() -> None:
    version_list = minecraft_launcher_lib.forge.list_forge_versions()
    assert isinstance(version_list[0], str)


def test_find_forge_version() -> None:
    assert isinstance(minecraft_launcher_lib.forge.find_forge_version("1.16.2"), str)
    assert minecraft_launcher_lib.forge.find_forge_version("Test123") is None


def test_is_forge_version_valid() -> None:
    assert minecraft_launcher_lib.forge.is_forge_version_valid("1.16.5-36.1.32") is True
    assert minecraft_launcher_lib.forge.is_forge_version_valid("Test123") is False


def test_supports_automatic_install() -> None:
    assert minecraft_launcher_lib.forge.supports_automatic_install("1.16.5-36.1.32") is True
    assert minecraft_launcher_lib.forge.supports_automatic_install("1.12-14.21.1.2443") is False
    assert minecraft_launcher_lib.forge.supports_automatic_install("invalid") is False


def test_forge_to_installed_version() -> None:
    assert minecraft_launcher_lib.forge.forge_to_installed_version("1.19-41.0.105") == "1.19-forge-41.0.105"

    with pytest.raises(ValueError):
        minecraft_launcher_lib.forge.forge_to_installed_version("Test123")



import minecraft_launcher_lib
import platform
import pathlib
import pytest
import shutil
import os


def _prepare_natives_test_env(tmp_path: pathlib.Path) -> None:
    shutil.copytree(os.path.join(os.path.dirname(__file__), "data", "versions"), tmp_path / "versions")


def test_extract_natives_linux(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    _prepare_natives_test_env(tmp_path)

    monkeypatch.setattr(platform, "system", lambda: "Linux")

    minecraft_launcher_lib.install.install_minecraft_version("natives", tmp_path)

    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "liblwjgl.so")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "liblwjgl64.so")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "libopenal.so")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "liblwjgl64.so")
    assert not os.path.isdir(tmp_path / "versions" / "natives" / "natives" / "META-INF")

    minecraft_launcher_lib.natives.extract_natives("natives", tmp_path, tmp_path / "extract")

    assert os.path.isfile(tmp_path / "extract" / "liblwjgl.so")
    assert os.path.isfile(tmp_path / "extract" / "liblwjgl64.so")
    assert os.path.isfile(tmp_path / "extract" / "libopenal.so")
    assert os.path.isfile(tmp_path / "extract" / "liblwjgl64.so")
    assert not os.path.isdir(tmp_path / "extract" / "META-INF")


def test_extract_natives_mac(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    _prepare_natives_test_env(tmp_path)

    monkeypatch.setattr(platform, "system", lambda: "Darwin")

    minecraft_launcher_lib.install.install_minecraft_version("natives", tmp_path)

    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "liblwjgl.jnilib")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "openal.dylib")
    assert not os.path.isdir(tmp_path / "versions" / "natives" / "natives" / "META-INF")

    minecraft_launcher_lib.natives.extract_natives("natives", tmp_path, tmp_path / "extract")

    assert os.path.isfile(tmp_path / "extract" / "liblwjgl.jnilib")
    assert os.path.isfile(tmp_path / "extract" / "openal.dylib")
    assert not os.path.isdir(tmp_path / "extract" / "META-INF")


def test_extract_natives_windows(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    _prepare_natives_test_env(tmp_path)

    monkeypatch.setattr(platform, "architecture", lambda: ("64bit",))
    monkeypatch.setattr(platform, "system", lambda: "Windows")

    minecraft_launcher_lib.install.install_minecraft_version("natives", tmp_path)

    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "lwjgl.dll")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "lwjgl64.dll")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "OpenAL32.dll")
    assert os.path.isfile(tmp_path / "versions" / "natives" / "natives" / "OpenAL64.dll")
    assert not os.path.isdir(tmp_path / "versions" / "natives" / "natives" / "META-INF")

    minecraft_launcher_lib.natives.extract_natives("natives", tmp_path, tmp_path / "extract64")

    assert os.path.isfile(tmp_path / "extract64" / "lwjgl.dll")
    assert os.path.isfile(tmp_path / "extract64" / "lwjgl64.dll")
    assert os.path.isfile(tmp_path / "extract64" / "OpenAL32.dll")
    assert os.path.isfile(tmp_path / "extract64" / "OpenAL64.dll")
    assert not os.path.isdir(tmp_path / "extract64" / "META-INF")

    monkeypatch.setattr(platform, "architecture", lambda: ("32bit",))

    minecraft_launcher_lib.install.install_minecraft_version("natives", tmp_path)
    minecraft_launcher_lib.natives.extract_natives("natives", tmp_path, tmp_path / "extract32")

    assert os.path.isfile(tmp_path / "extract32" / "lwjgl.dll")
    assert os.path.isfile(tmp_path / "extract32" / "lwjgl64.dll")
    assert os.path.isfile(tmp_path / "extract32" / "OpenAL32.dll")
    assert os.path.isfile(tmp_path / "extract32" / "OpenAL64.dll")
    assert not os.path.isdir(tmp_path / "extract64" / "META-INF")


def test_extract_natives_invalid_version(tmp_path: pathlib.Path) -> None:
    # Checks if the VersionNotFound exception raised
    with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
        minecraft_launcher_lib.natives.extract_natives("InvalidVersion", str(tmp_path), tmp_path / "extract")



from typing import Dict, Literal, TypedDict


class VanillaLauncherProfilesJsonProfile(TypedDict, total=False):
    created: str
    gameDir: str
    icon: str
    javaArgs: str
    javaDir: str
    lastUsed: str
    lastVersionId: str
    name: str
    resolution: Dict[Literal["height", "width"], int]
    type: str


class VanillaLauncherProfilesJson(TypedDict):
    profiles: Dict[str, VanillaLauncherProfilesJsonProfile]
    version: int



from ._test_helper import prepare_test_versions, prepare_requests_mock
import minecraft_launcher_lib
import requests_mock
import platform
import hashlib
import pathlib
import pytest
import os


def _assert_downloaded_file(path: pathlib.Path, size: int, hash: str) -> None:
    assert os.path.isfile(path)

    assert os.path.getsize(path) == size

    with open(path, "rb") as f:
        assert hashlib.sha1(f.read()).hexdigest() == hash


def test_install_minecraft_version(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)
    prepare_requests_mock(requests_mock)

    minecraft_launcher_lib.install.install_minecraft_version("test1", tmp_path)

    _assert_downloaded_file(tmp_path / "libraries" / "ca" / "weblite" / "java-objc-bridge" / "1.1" / "java-objc-bridge-1.1.jar", 25, "2f8a26755f00bc8dc54cbf5e613f6d00f02a099e")
    _assert_downloaded_file(tmp_path / "libraries" / "org" / "slf4j" / "slf4j-api" / "2.0.1" / "slf4j-api-2.0.1.jar", 20, "03b7e695aa5cd7ca418ee6710e304d64d0491929")
    _assert_downloaded_file(tmp_path / "assets" / "log_configs" / "client-1.12.xml", 888, "bd65e7d2e3c237be76cfbef4c2405033d7f91521")
    _assert_downloaded_file(tmp_path / "versions" / "test1" / "test1.jar", 7, "173219a75174abb3e3a7bfd36148129df03f9123")


def test_install_minecraft_version_assets(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)
    prepare_requests_mock(requests_mock)

    minecraft_launcher_lib.install.install_minecraft_version("assets", tmp_path)

    assert (tmp_path / "assets" / "objects" / "86" / "86f7e437faa5a7fce15d1ddcb9eaeaea377667b8").is_file()
    assert (tmp_path / "assets" / "objects" / "e9" / "e9d71f5ee7c92d6dc9e92ffdad17b8bd49418f98").is_file()
    assert (tmp_path / "assets" / "objects" / "84" / "84a516841ba77a5b4648de2cd0dfcb30ea46dbb4").is_file()
    assert os.listdir(tmp_path / "assets" / "log_configs") == ["client-1.12.xml"]
    assert os.listdir(tmp_path / "assets" / "indexes") == ["test.json"]
    assert len(os.listdir(tmp_path / "assets" / "objects")) == 3


def test_install_minecraft_version_runtime(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)
    prepare_requests_mock(requests_mock)

    minecraft_launcher_lib.install.install_minecraft_version("runtime", tmp_path)

    assert (tmp_path / "runtime" / "java-runtime-test" / "linux" / ".version").read_text().strip() == "17.0.3"


def test_install_minecraft_version_inherit(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)
    prepare_requests_mock(requests_mock)

    minecraft_launcher_lib.install.install_minecraft_version("inherit", tmp_path)

    _assert_downloaded_file(tmp_path / "libraries" / "ca" / "weblite" / "java-objc-bridge" / "1.1" / "java-objc-bridge-1.1.jar", 25, "2f8a26755f00bc8dc54cbf5e613f6d00f02a099e")
    _assert_downloaded_file(tmp_path / "libraries" / "org" / "slf4j" / "slf4j-api" / "2.0.1" / "slf4j-api-2.0.1.jar", 20, "03b7e695aa5cd7ca418ee6710e304d64d0491929")
    _assert_downloaded_file(tmp_path / "libraries" / "com" / "ibm" / "icu" / "icu4j" / "71.1" / "icu4j-71.1.jar", 14, "78ef2a15d1afad3e435cdcaf12c502ac290ca707")
    _assert_downloaded_file(tmp_path / "assets" / "log_configs" / "client-1.12.xml", 888, "bd65e7d2e3c237be76cfbef4c2405033d7f91521")
    _assert_downloaded_file(tmp_path / "versions" / "inherit" / "inherit.jar", 7, "173219a75174abb3e3a7bfd36148129df03f9123")
    _assert_downloaded_file(tmp_path / "versions" / "test1" / "test1.jar", 7, "173219a75174abb3e3a7bfd36148129df03f9123")


def test_install_minecraft_version_extended(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)

    prepare_requests_mock(requests_mock)

    minecraft_launcher_lib.install.install_minecraft_version("test2", tmp_path)

    _assert_downloaded_file(tmp_path / "libraries" / "ca" / "weblite" / "java-objc-bridge" / "1.1" / "java-objc-bridge-1.1.jar", 1, "86f7e437faa5a7fce15d1ddcb9eaeaea377667b8")


def test_install_minecraft_version_invalid_checksum(monkeypatch: pytest.MonkeyPatch, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)

    prepare_requests_mock(requests_mock)

    with pytest.raises(minecraft_launcher_lib.exceptions.InvalidChecksum):
        minecraft_launcher_lib.install.install_minecraft_version("checksum", tmp_path)


def test_install_minecraft_version_invalid_version(tmp_path: pathlib.Path) -> None:
    # Checks if the VersionNotFound exception raised
    with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
        minecraft_launcher_lib.install.install_minecraft_version("InvalidVersion", str(tmp_path))



import minecraft_launcher_lib
from typing import List
import subprocess
import platform
import pathlib
import pytest
import sys


_org_subprocess_run = subprocess.run


def _subprocess_mock(cmd: List[str], **kwargs: dict) -> subprocess.CompletedProcess:
    return _org_subprocess_run([sys.executable, pathlib.Path(__file__).parent / "data" / "java_utils" / "java_mock.py"], **kwargs)


def test_get_java_information_unix(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(subprocess, "run", _subprocess_mock)

    with pytest.raises(ValueError, match="java was not found"):
        minecraft_launcher_lib.java_utils.get_java_information(tmp_path)

    (tmp_path / "bin").mkdir()
    (tmp_path / "bin" / "java").touch()

    info = minecraft_launcher_lib.java_utils.get_java_information(tmp_path)

    assert info["path"] == str(tmp_path)
    assert info["name"] == tmp_path.name
    assert info["version"] == "19.0.1"
    assert info["is_64bit"] is True
    assert info["openjdk"] is True
    assert info["java_path"] == str(tmp_path / "bin" / "java")
    assert info["javaw_path"] is None


def test_get_java_information_windows(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Windows")
    monkeypatch.setattr(subprocess, "run", _subprocess_mock)

    with pytest.raises(ValueError, match="java.exe was not found"):
        minecraft_launcher_lib.java_utils.get_java_information(tmp_path)

    (tmp_path / "bin").mkdir()
    (tmp_path / "bin" / "java.exe").touch()

    info = minecraft_launcher_lib.java_utils.get_java_information(tmp_path)

    assert info["path"] == str(tmp_path)
    assert info["name"] == tmp_path.name
    assert info["version"] == "19.0.1"
    assert info["is_64bit"] is True
    assert info["openjdk"] is True
    assert info["java_path"] == str(tmp_path / "bin" / "java.exe")
    assert info["javaw_path"] == str(tmp_path / "bin" / "javaw.exe")


def test_find_system_java_versions(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Windows")

    minecraft_launcher_lib.java_utils.find_system_java_versions(additional_directories=[tmp_path])

    monkeypatch.setattr(platform, "system", lambda: "Linux")

    (tmp_path / "test.txt").touch()
    (tmp_path / "testjava").mkdir()
    (tmp_path / "testjava" / "bin").mkdir()
    (tmp_path / "testjava" / "bin" / "java").touch()

    version_list = minecraft_launcher_lib.java_utils.find_system_java_versions(additional_directories=[tmp_path])

    assert isinstance(version_list, list)

    for i in version_list:
        assert isinstance(i, str)

    assert str(tmp_path / "testjava") in version_list


def test_find_system_java_versions_information(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(subprocess, "run", _subprocess_mock)

    (tmp_path / "testjava").mkdir()
    (tmp_path / "testjava" / "bin").mkdir()
    (tmp_path / "testjava" / "bin" / "java").touch()

    minecraft_launcher_lib.java_utils.find_system_java_versions_information(additional_directories=[tmp_path])



from ._test_helper import prepare_test_versions, get_test_callbacks, prepare_requests_mock
import minecraft_launcher_lib
import pytest_subtests
import requests_mock
import subprocess
import platform
import pathlib
import pytest
import shutil


def test_quilt_get_all_minecraft_versions(requests_mock: requests_mock.Mocker) -> None:
    prepare_requests_mock(requests_mock)

    version_list = minecraft_launcher_lib.quilt.get_all_minecraft_versions()
    for i in version_list:
        assert isinstance(i["version"], str)
        assert isinstance(i["stable"], bool)


def test_quilt_get_stable_minecraft_versions(requests_mock: requests_mock.Mocker) -> None:
    prepare_requests_mock(requests_mock)

    version_list = minecraft_launcher_lib.quilt.get_stable_minecraft_versions()
    assert isinstance(version_list[0], str)


def test_quilt_get_latest_minecraft_version(requests_mock: requests_mock.Mocker) -> None:
    prepare_requests_mock(requests_mock)

    assert minecraft_launcher_lib.quilt.get_latest_minecraft_version() == "unstable"


def test_quilt_get_latest_stable_minecraft_version(requests_mock: requests_mock.Mocker) -> None:
    prepare_requests_mock(requests_mock)

    assert minecraft_launcher_lib.quilt.get_latest_stable_minecraft_version() == "test2"


def test_quilt_is_minecraft_version_supported(requests_mock: requests_mock.Mocker) -> None:
    prepare_requests_mock(requests_mock)

    assert minecraft_launcher_lib.quilt.is_minecraft_version_supported("test2") is True
    assert minecraft_launcher_lib.quilt.is_minecraft_version_supported("test1") is False


def test_quilt_get_all_loader_versions() -> None:
    version_list = minecraft_launcher_lib.quilt.get_all_loader_versions()
    for i in version_list:
        assert isinstance(i["separator"], str)
        assert isinstance(i["build"], int)
        assert isinstance(i["maven"], str)
        assert isinstance(i["version"], str)


def test_quilt_get_latest_loader_version() -> None:
    assert isinstance(minecraft_launcher_lib.quilt.get_latest_loader_version(), str)


def test_quilt_get_latest_installer_version() -> None:
    assert isinstance(minecraft_launcher_lib.quilt.get_latest_installer_version(), str)


def test_install_quilt(monkeypatch: pytest.MonkeyPatch, subtests: pytest_subtests.SubTests, requests_mock: requests_mock.Mocker, tmp_path: pathlib.Path) -> None:
    monkeypatch.setattr(minecraft_launcher_lib.quilt, "is_minecraft_version_supported", lambda version: True if version == "test1" else False)
    monkeypatch.setattr(minecraft_launcher_lib.quilt, "get_latest_installer_version", lambda: "testinstaller")
    monkeypatch.setattr(minecraft_launcher_lib.quilt, "get_latest_loader_version", lambda: "testloader")
    monkeypatch.setattr(subprocess, "run", lambda cmd, **kwargs: subprocess.CompletedProcess([], 0))
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    requests_mock.get("https://maven.quiltmc.org/repository/release/org/quiltmc/quilt-installer/testinstaller/quilt-installer-testinstaller.jar")
    requests_mock.real_http = True

    prepare_test_versions(tmp_path)
    prepare_requests_mock(requests_mock)

    shutil.copytree(tmp_path / "versions" / "test1", tmp_path / "versions" / "quilt-loader-testloader-test1")
    (tmp_path / "versions" / "quilt-loader-testloader-test1" / "test1.json").rename(tmp_path / "versions" / "quilt-loader-testloader-test1" / "quilt-loader-testloader-test1.json")

    with subtests.test("Install"):
        minecraft_launcher_lib.quilt.install_quilt("test1", tmp_path, callback=get_test_callbacks())

    with subtests.test("ExternalProgramError"):
        monkeypatch.setattr(subprocess, "run", lambda cmd, **kwargs: subprocess.CompletedProcess([], 1))

        with pytest.raises(minecraft_launcher_lib.exceptions.ExternalProgramError):
            minecraft_launcher_lib.quilt.install_quilt("test1", tmp_path)

    with subtests.test("UnsupportedVersion"):
        with pytest.raises(minecraft_launcher_lib.exceptions.UnsupportedVersion):
            minecraft_launcher_lib.quilt.install_quilt("natives", tmp_path)

    with subtests.test("VersionNotFound"):
        with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
            minecraft_launcher_lib.quilt.install_quilt("invalid", tmp_path)



from ._test_helper import prepare_test_versions
import minecraft_launcher_lib
from typing import List
import platform
import pathlib
import pytest
import os


def _check_argument(command: List[str], argument: str, value: str) -> bool:
    # Checks if a argument has the given value
    try:
        index = command.index(argument)
    except Exception:
        return False
    try:
        return command[index + 1] == value
    except Exception:
        return False


def test_get_minecraft_command_executable_basic(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)
    options = {}
    options["username"] = "Player"
    options["uuid"] = "Testuuid"
    options["token"] = "testtoken"
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--username", "Player")
    assert _check_argument(command, "--uuid", "Testuuid")
    assert _check_argument(command, "--accessToken", "testtoken")


def test_get_minecraft_command_java(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    # Test with option
    options = {"executablePath": "test"}
    command = minecraft_launcher_lib.command.get_minecraft_command("java", str(tmp_path), options)
    assert command[0] == "test"

    # Test without option
    command = minecraft_launcher_lib.command.get_minecraft_command("java", str(tmp_path), {})
    assert command[0] == "java"

    # Test if using the runtime path

    java_path = (tmp_path / "runtime" / "java-runtime-test" / "linux" / "java-runtime-test" / "bin" / "java")
    os.makedirs(java_path.parent)
    java_path.touch()
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    command = minecraft_launcher_lib.command.get_minecraft_command("java", str(tmp_path), {})
    assert command[0] == str(java_path)


def test_get_minecraft_command_jvm_arguments(tmp_path: pathlib.Path):
    prepare_test_versions(tmp_path)
    options = {"jvmArguments": ["a", "b"]}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert command[1] == "a"
    assert command[2] == "b"


def test_get_minecraft_command_game_directory(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    # Test with option
    options = {"gameDirectory": "gamedir"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--gameDir", "gamedir")

    # Test without option
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), {})
    assert _check_argument(command, "--gameDir", str(tmp_path))


def test_get_minecraft_command_demo(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)
    # Test with option
    options = {"demo": True}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--demo" in command
    # Test without option
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), {})
    assert "--demo" not in command


def test_get_minecraft_command_custom_resolution(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    # Test default values of width and height
    options = {"customResolution": True}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--width", "854")
    assert _check_argument(command, "--height", "480")

    # Test custom values of width and height
    options = {}
    options["customResolution"] = True
    options["resolutionWidth"] = "1000"
    options["resolutionHeight"] = "500"
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--width", "1000")
    assert _check_argument(command, "--height", "500")

    # Test without options
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), {})
    assert "--width" not in command
    assert "--height" not in command


def test_get_minecraft_command_natives_directory(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)
    # Test with option
    options = {"nativesDirectory": "test"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "-Djava.library.path=test" in command
    # Test without option
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), {})
    default_natives_path = os.path.join(tmp_path, "versions", "test1", "natives")
    assert f"-Djava.library.path={default_natives_path}"


def test_get_minecraft_command_logging_config(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    options = {"enableLoggingConfig": True}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "-Dlog4j.configurationFile=" + str(tmp_path / "assets" / "log_configs" / "client-1.12.xml") in command


def test_get_minecraft_command_server_port(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    options = {"server": "example.com"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--server", "example.com")
    assert "--port" not in command

    options["port"] = "123"
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--server", "example.com")
    assert _check_argument(command, "--port", "123")


def test_get_minecraft_command_disable_multiplayer_chat(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    options = {"disableMultiplayer": True}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--disableMultiplayer" in command
    assert "--disableChat" not in command

    options = {"disableChat": True}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--disableMultiplayer" not in command
    assert "--disableChat" in command


def test_get_minecraft_command_quick_play(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    options = {}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--quickPlayPath" not in command
    assert "--quickPlaySingleplayer" not in command
    assert "--quickPlayMultiplayer" not in command
    assert "--quickPlayRealms" not in command

    options = {"quickPlayPath": "testPath"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert _check_argument(command, "--quickPlayPath", "testPath")
    assert "--quickPlaySingleplayer" not in command
    assert "--quickPlayMultiplayer" not in command
    assert "--quickPlayRealms" not in command

    options = {"quickPlaySingleplayer": "testSingleplayer"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--quickPlayPath" not in command
    assert _check_argument(command, "--quickPlaySingleplayer", "testSingleplayer")
    assert "--quickPlayMultiplayer" not in command
    assert "--quickPlayRealms" not in command

    options = {"quickPlayMultiplayer": "testMultiplayer"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--quickPlayPath" not in command
    assert "--quickPlaySingleplayer" not in command
    assert _check_argument(command, "--quickPlayMultiplayer", "testMultiplayer")
    assert "--quickPlayRealms" not in command

    options = {"quickPlayRealms": "testRealms"}
    command = minecraft_launcher_lib.command.get_minecraft_command("test1", str(tmp_path), options)
    assert "--quickPlayPath" not in command
    assert "--quickPlaySingleplayer" not in command
    assert "--quickPlayMultiplayer" not in command
    assert _check_argument(command, "--quickPlayRealms", "testRealms")


def test_get_minecraft_command_old_arguments(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    options = {}
    command = minecraft_launcher_lib.command.get_minecraft_command("natives", str(tmp_path), options)
    assert "--width" not in command
    assert "--height" not in command
    assert "--demo" not in command

    options = {"customResolution": True, "resolutionWidth": "1000", "resolutionHeight": "500"}
    command = minecraft_launcher_lib.command.get_minecraft_command("natives", str(tmp_path), options)
    assert _check_argument(command, "--width", "1000")
    assert _check_argument(command, "--height", "500")
    assert "--demo" not in command

    options = {"demo": True}
    command = minecraft_launcher_lib.command.get_minecraft_command("natives", str(tmp_path), options)
    assert "--width" not in command
    assert "--height" not in command
    assert "--demo" in command


def test_get_minecraft_command_inherit(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)

    options = {}
    minecraft_launcher_lib.command.get_minecraft_command("inherit", str(tmp_path), options)


def test_get_minecraft_command_extended(tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(platform, "system", lambda: "Linux")

    prepare_test_versions(tmp_path)

    minecraft_launcher_lib.command.get_minecraft_command("test2", str(tmp_path), {})


def test_get_minecraft_command_invalid_version(tmp_path: pathlib.Path) -> None:
    # Checks if the VersionNotFound exception raised
    with pytest.raises(minecraft_launcher_lib.exceptions.VersionNotFound):
        minecraft_launcher_lib.command.get_minecraft_command("InvalidVersion", str(tmp_path), {})


def test_get_minecraft_command_with_path(tmp_path: pathlib.Path) -> None:
    prepare_test_versions(tmp_path)
    # Test if get_minecraft_command works with os.PathLike
    minecraft_launcher_lib.command.get_minecraft_command("test1", tmp_path, {})



from typing import List, TypedDict
import requests
import datetime


class RequestsResponseCache(TypedDict):
    response: requests.models.Response
    datetime: datetime.datetime


class MavenMetadata(TypedDict):
    release: str
    latest: str
    versions: List[str]



"mrpack allows you to install Modpacks from the `Mrpack Format <https://docs.modrinth.com/docs/modpacks/format_definition>`_"
from ._helper import download_file, empty, check_path_inside_minecraft_directory
from .types import MrpackInformation, MrpackInstallOptions, CallbackDict
from ._internal_types.mrpack_types import MrpackIndex, MrpackFile
from .install import install_minecraft_version
from typing import List, Union, Optional
from .forge import install_forge_version
from .fabric import install_fabric
from .quilt import install_quilt
import zipfile
import json
import os


def _filter_mrpack_files(file_list: List[MrpackFile], mrpack_install_options: MrpackInstallOptions) -> List[MrpackFile]:
    """
    Gets all Mrpack Files that should be installed
    """
    filtered_list: List[MrpackFile] = []
    for file in file_list:
        if "env" not in file:
            filtered_list.append(file)
            continue

        if file["env"]["client"] == "required":
            filtered_list.append(file)
        if file["env"]["client"] == "optional" and file["path"] in mrpack_install_options.get("optionalFiles", []):
            filtered_list.append(file)

    return filtered_list


def get_mrpack_information(path: Union[str, os.PathLike]) -> MrpackInformation:
    """
    Gets some Information from a .mrpack file

    :param path: The Path the the .mrpack file
    :type path: Union[str, os.PathLike]
    :return: The Information about the given Mrpack
    """
    with zipfile.ZipFile(path, "r") as zf:
        with zf.open("modrinth.index.json", "r") as f:
            index: MrpackIndex = json.load(f)

            information: MrpackInformation = {}  # type: ignore
            information["name"] = index["name"]
            information["summary"] = index.get("summary", "")
            information["versionId"] = index["versionId"]
            information["formatVersion"] = index["formatVersion"]
            information["minecraftVersion"] = index["dependencies"]["minecraft"]

            information["optionalFiles"] = []
            for file in index["files"]:
                if "env" not in file:
                    continue

                if file["env"]["client"] == "optional":
                    information["optionalFiles"].append(file["path"])

            return information


def install_mrpack(path: Union[str, os.PathLike], minecraft_directory: Union[str, os.PathLike], modpack_directory: Optional[Union[str, os.PathLike]] = None, callback: Optional[CallbackDict] = None, mrpack_install_options: Optional[MrpackInstallOptions] = None) -> None:
    """
    Installs a .mrpack file

    :param path: The Path the the .mrpack file
    :param minecraft_directory: he path to your Minecraft directory
    :param modpack_directory: If you want to install the Pack in another Directory than your Minecraft directory, set it here.
    :param callback: The same dict as for :func:`~minecraft_launcher_lib.install.install_minecraft_version`
    :param mrpack_install_options: Some Options to install the Pack (see below)
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory

    ``mrpack_install_options`` is a dict. All Options are Optional.

    .. code:: python

        mrpack_install_options = {
            "optionalFiles": [], # List with all Optional files
            "skipDependenciesInstall": False # If you want to skip the Dependencie install. Only used for testing purposes.
        }
    """
    minecraft_directory = os.path.abspath(minecraft_directory)
    path = os.path.abspath(path)

    if modpack_directory is None:
        modpack_directory = minecraft_directory
    else:
        modpack_directory = os.path.abspath(modpack_directory)

    if callback is None:
        callback = {}

    if mrpack_install_options is None:
        mrpack_install_options = {}

    with zipfile.ZipFile(path, "r") as zf:
        with zf.open("modrinth.index.json", "r") as f:
            index: MrpackIndex = json.load(f)

        # Download the files
        callback.get("setStatus", empty)("Download mrpack files")
        file_list = _filter_mrpack_files(index["files"], mrpack_install_options)
        callback.get("setMax", empty)(len(file_list))
        for count, file in enumerate(file_list):
            full_path = os.path.abspath(os.path.join(modpack_directory, file["path"]))

            check_path_inside_minecraft_directory(modpack_directory, full_path)

            download_file(file["downloads"][0], full_path, sha1=file["hashes"]["sha1"], callback=callback)

            callback.get("setProgress", empty)(count + 1)

        # Extract the overrides
        callback.get("setStatus", empty)("Extract overrides")
        for zip_name in zf.namelist():
            # Check if the entry is in the overrides and if it is a file
            if (not zip_name.startswith("overrides/") and not zip_name.startswith("client-overrides/")) or zf.getinfo(zip_name).file_size == 0:
                continue

            # Remove the overrides at the start of the Name
            # We don't have removeprefix() in Python 3.8
            if zip_name.startswith("client-overrides/"):
                file_name = zip_name[len("client-overrides/"):]
            else:
                file_name = zip_name[len("overrides/"):]

            # Constructs the full Path
            full_path = os.path.abspath(os.path.join(modpack_directory, file_name))

            check_path_inside_minecraft_directory(modpack_directory, full_path)

            callback.get("setStatus", empty)(f"Extract {zip_name}]")
            zf.extract(zip_name, full_path)

        if mrpack_install_options.get("skipDependenciesInstall"):
            return

        # Install dependencies
        callback.get("setStatus", empty)("Installing Minecraft " + index["dependencies"]["minecraft"])
        install_minecraft_version(index["dependencies"]["minecraft"], minecraft_directory, callback=callback)

        if "forge" in index["dependencies"]:
            forge_version = index["dependencies"]["minecraft"] + "-" + index["dependencies"]["forge"]
            callback.get("setStatus", empty)(f"Installing Forge {forge_version}")
            install_forge_version(forge_version, minecraft_directory, callback=callback)

        if "fabric-loader" in index["dependencies"]:
            callback.get("setStatus", empty)("Installing Fabric " + index["dependencies"]["fabric-loader"] + " for Minecraft " + index["dependencies"]["minecraft"])
            install_fabric(index["dependencies"]["minecraft"], minecraft_directory, loader_version=index["dependencies"]["fabric-loader"], callback=callback)

        if "quilt-loader" in index["dependencies"]:
            callback.get("setStatus", empty)("Installing Quilt " + index["dependencies"]["quilt-loader"] + " for Minecraft " + index["dependencies"]["minecraft"])
            install_quilt(index["dependencies"]["minecraft"], minecraft_directory, loader_version=index["dependencies"]["quilt-loader"], callback=callback)


def get_mrpack_launch_version(path: Union[str, os.PathLike]) -> str:
    """
    Returns that Version that needs to be used with :func:`~minecraft_launcher_lib.command.get_minecraft_command`.

    :param path: The Path the the .mrpack file
    :return: The version
    """
    with zipfile.ZipFile(path, "r") as zf:
        with zf.open("modrinth.index.json", "r") as f:
            index: MrpackIndex = json.load(f)

            if "forge" in index["dependencies"]:
                return index["dependencies"]["minecraft"] + "-forge-" + index["dependencies"]["forge"]
            elif "fabric-loader" in index["dependencies"]:
                return "fabric-loader-" + index["dependencies"]["fabric-loader"] + "-" + index["dependencies"]["minecraft"]
            elif "quilt-loader" in index["dependencies"]:
                return "quilt-loader-" + index["dependencies"]["quilt-loader"] + "-" + index["dependencies"]["minecraft"]
            else:
                return index["dependencies"]["minecraft"]



"natives contains a function for extracting natives libraries to a specific folder"
from ._helper import parse_rule_list, check_path_inside_minecraft_directory
from ._internal_types.shared_types import ClientJson, ClientJsonLibrary
from typing import List, Dict, Literal, Union
from .exceptions import VersionNotFound
import platform
import zipfile
import json
import os

__all__ = ["extract_natives"]


def get_natives(data: ClientJsonLibrary) -> str:
    """
    Returns the native part from the json data
    """
    if platform.architecture()[0] == "32bit":
        arch_type = "32"
    else:
        arch_type = "64"

    if "natives" in data:
        if platform.system() == 'Windows':
            if "windows" in data["natives"]:
                return data["natives"]["windows"].replace("${arch}", arch_type)
            else:
                return ""
        elif platform.system() == 'Darwin':
            if "osx" in data["natives"]:
                return data["natives"]["osx"].replace("${arch}", arch_type)
            else:
                return ""
        else:
            if "linux" in data["natives"]:
                return data["natives"]["linux"].replace("${arch}", arch_type)
            else:
                return ""
    else:
        return ""


def extract_natives_file(filename: str, extract_path: str, extract_data: Dict[Literal["exclude"], List[str]]) -> None:
    """
    Unpack natives
    """
    try:
        os.mkdir(extract_path)
    except Exception:
        pass

    with zipfile.ZipFile(filename, "r") as zf:
        for i in zf.namelist():
            for e in extract_data["exclude"]:
                if i.startswith(e):
                    break
            else:
                zf.extract(i, extract_path)


def extract_natives(versionid: str, path: Union[str, os.PathLike], extract_path: str) -> None:
    """
    Extract all native libraries from a version into the given directory. The directory will be created, if it does not exist.

    :param version: The Minecraft version
    :param minecraft_directory: The path to your Minecraft directory
    :param callback: The same dict as for :func:`~minecraft_launcher_lib.install.install_minecraft_version`
    :raises VersionNotFound: The Minecraft version was not found
    :raises FileOutsideMinecraftDirectory: A File should be placed outside the given Minecraft directory

    The natives are all extracted while installing. So you don't need to use this function in most cases.
    """
    if not os.path.isfile(os.path.join(path, "versions", versionid, versionid + ".json")):
        raise VersionNotFound(versionid)

    with open(os.path.join(path, "versions", versionid, versionid + ".json"), "r", encoding="utf-8") as f:
        data: ClientJson = json.load(f)

    for i in data["libraries"]:
        # Check, if the rules allow this lib for the current system
        if "rules" in i and not parse_rule_list(i["rules"], {}):
            continue

        current_path = os.path.join(path, "libraries")
        lib_path, name, version = i["name"].split(":")

        for lib_part in lib_path.split("."):
            current_path = os.path.join(current_path, lib_part)

        current_path = os.path.join(current_path, name, version)
        native = get_natives(i)

        if native == "":
            continue

        jar_filename_native = name + "-" + version + "-" + native + ".jar"
        if "extract" in i:
            check_path_inside_minecraft_directory(path, extract_path)
            extract_natives_file(os.path.join(current_path, jar_filename_native), extract_path, i["extract"])
